/*
 *  This file is part of DUMPI: 
 *               The MPI profiling library from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <dumpi/libdumpi/callprofile.h>

#ifdef DUMPI_ENABLE_INSTRUMENTATION

#include <dumpi/libdumpi/callprofile-addrset.h>
#include <dumpi/libdumpi/mpibindings.h>
#include <assert.h>
#include <stdint.h>
#include <stdio.h>

static int calldepth = 0;

/* Indicate entry into a profiled function. */
void __cyg_profile_func_enter(void *fn, void *site) {
  dumpi_func_call stat;
  uint16_t thread;
  dumpi_time cpu, wall;
  /*
  printf("__cyg_profile_func_enter(%p [\"%s\"], %p)\n",
	 fn, dumpi_function_name(fn), site);
  */
  if(dumpi_global == NULL)
    libdumpi_init();
  STARTTIME(cpu, wall);
  THREAD_ID(thread);
  (void)site;
  ++calldepth;
  insert_address(fn);
  if(dumpi_global->output->function[DUMPI_Function_enter]) {
    stat.fn = (uint64_t)fn;
    STOPTIME(cpu, wall);
    dumpi_write_func_enter(&stat, thread, &cpu, &wall, dumpi_global->perf,
			   dumpi_global->output, dumpi_global->profile);
  }
}

/* Indicate exit from a profiled function. */
void __cyg_profile_func_exit(void *fn, void *site) {
  dumpi_func_call stat;
  uint16_t thread;
  dumpi_time cpu, wall;
  if(dumpi_global == NULL)
    libdumpi_init();
  STARTTIME(cpu, wall);
  THREAD_ID(thread);
  (void)site;
  --calldepth;
  insert_address(fn); /* should not be needed - but better safe than sorry */
  if(dumpi_global->output->function[DUMPI_Function_exit]) {
    stat.fn = (uint64_t)fn;
    STOPTIME(cpu, wall);
    dumpi_write_func_exit(&stat, thread, &cpu, &wall, dumpi_global->perf,
			  dumpi_global->output, dumpi_global->profile);
  }
}

#ifdef DUMPI_DLADDR_WORKS
#define _GNU_SOURCE 1
#define __USE_GNU 1
#include <dlfcn.h>

const char* dumpi_function_name(void *fn) {
  Dl_info din;
  assert(dladdr(fn, &din) != 0);
  return din.dli_sname;
}

#else /* ! DUMPI_DLADDR_WORKS */

static const char *none = "(none)";

const char* dumpi_function_name(void *fn) {
  (void)fn;
  return none;
}

#endif /* DUMPI_DLADDR_WORKS */

/*
 * Get the number of unique function addresses profiled so far.
 */
int dumpi_profile_unique_address_count(void) {
  dumpi_addr_iterator it;
  int retval = 0;
  for(it = begin_iterator(); valid_iterator(&it); increment_iterator(&it)) {
    ++retval;
  }
  return retval;
}
  
/*
 * Get a listing of all unique function addresses profiled so far.
 */
uint64_t* dumpi_profile_unique_addresses(void) {
  int curr;
  uint64_t *addresses;
  dumpi_addr_iterator it;
  int arrsize = dumpi_profile_unique_address_count();
  addresses = (uint64_t*)calloc(arrsize+1, sizeof(uint64_t));
  assert(addresses != NULL);
  curr = 0;
  for(it = begin_iterator(); valid_iterator(&it); increment_iterator(&it)) {
    assert(curr < arrsize);
    addresses[curr] = (uint64_t)iterator_value(&it);
    ++curr;
  }
  return addresses;
}

#else

#include <stdint.h>
#include <stdlib.h>

/*
 * Get the number of unique function addresses profiled so far.
 */
int dumpi_profile_unique_address_count(void) {
  return 0;
}
  
/*
 * Get a listing of all unique function addresses profiled so far.
 */
uint64_t* dumpi_profile_unique_addresses(void) {
  return NULL;
}

static const char *none = "(none)";

const char* dumpi_function_name(void *fn) {
  (void)fn;
  return none;
}

#endif /* ! DUMPI_ENABLE_INSTRUMENTATION */

