/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_MACROBRIDGE_MACRORECVER_H
#define SSTMAC_MACROBRIDGE_MACRORECVER_H

#include <sst/sst_config.h>
#include <sst/core/component.h>
#include <sst/core/eventFunctor.h>
#include <sst/core/link.h>

namespace sstmac {

  /**
   * A bridge component to instantiate a bridge between SST/macroscale
   * and SST/microscale.  The bridge will allow SST/macroscale to run
   * over the SST/microscale event manager and (eventually) transfer
   * events between the two scales.
   *
   * Since the SST/microscale implementation doesn't allow a component to
   * recv events to itself, we provide two interfaces for the
   * event handler.  This is the recver (all event completion arrives here).
   *
   * The macrorecver does a much less impressive job than the macrosender
   * (which actually serves to run the entire simulation) -- this type
   * simply provides a little bit of sanity (and actual memory management -- 
   * which may seem like an alien concept to some people).
   */
  class macrorecver : public SST::Component {
    /// The port wiring this component to others.
    /// This component will never send on its port, but we can't get
    /// wired without it.
    SST::Link *port_;

  private:
    /// Copy constructor is blocked so we can identify possible copy issues.
    macrorecver(const macrorecver&) : 
      Component(*this)
    {}
  
    /// Assignment operator is blocked to identify possible copy issues.
    void operator=(const macrorecver&) {}

    /// Private method to forward incoming component events.
    void handle_events(SST::Event *evt);

    /// Private method to set up the wiring between components.
    void get_wired(const std::string &target);

  public:
    /// This constructor is called by the SST runtime framework.
    macrorecver(SST::ComponentId_t id,
                SST::Component::Params_t &params);

    virtual ~macrorecver();

    /// As far as I can tell, this method gets called by the runtime system
    /// after setup is complete (well that sure makes a lot of sense).
    int Setup();

    /// This should be what happens when we're all done
    /// (although you never can tell with some people).
    int Finish();
  };


} // end of namespace sstmac.

#endif // ! SSTMAC_MACROBRIDGE_MACRORECVER_H
