/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */


// hdtorus.cc: Implementation of a high dimensional torus network.
//
// Author: Ali Pinar <apinar@sandia.gov> 
// 
 
#include <sstream>
#include <sstmac/network/hdtorus.h>
// #include <sstmac/network/ring.h>  
#include <stdio.h> 
 
using std::set;
using std::vector;
using std::string;

namespace sstmac {
        
// ------------------------------------------------------------------------------------------------------
/*  Computes the index array for all vertices  for a high dimensional torus 
    It uses dsize : the number of dimensions
        dimensions_:  an array of size dsize that stores each dimension 
*/

void hdtorus::compute_vertex_numbers(std::vector <std::vector<int> > &vertices, int P)
 {
   int bsize=P;
   
   int dsize=dimensions_.size();
   /* Below bsize  denotes how many consecutive  processors will be numbered 
      the same in the  d-tg digit.  Note that  the for loop with  index i
      does not necessarily run P  times, since i is incremented each time  
      in the innermost loop 
    */
   for (int d=0;d<dsize; d++) {
     bsize=bsize/dimensions_[d];   
     for (int i=0; i<P;) 
       for(int k=0;k<dimensions_[d];k++)
         for(int j=0;j<bsize;j++)
           vertices[i++][d]=k;   
   }  
 }
  
// ------------------------------------------------------------------------------------------------------
// Generates  the vertex numbers --and the vertices-- for a given number 
// of processors, P
//
 
  void hdtorus::compute_vertices(std::vector <std::vector<int> > &vertices, set<vertex> &vertices_set, int id)
  {  
    int i,P;
    int dsize=dimensions_.size();

    if (!vertices.size()) { 
      std::cout << " empty vertices in compute_vertices \n";
      std::cout.flush();
      for(i=0,P=1;i<dsize;P*=dimensions_[i++]);
    }
    P=vertices.size();
    
    if (id <0) /* single processor at each node */
      for(int i=0; i<P;i++)
        vertices_set.insert(vertex(vertices[i]));
    else   /* multiple processors at each node */
      for(int i=0; i<P;i++){
        vertices[i][dsize]=id;
        vertices_set.insert(vertex(vertices[i]));
      }
  }
  
  // --------------------------------------------------------------------------------------------------------
  // Generates the edges 
  
  void  hdtorus::compute_switch_edges(std::vector <std::vector<int> > &vertices, set<edge> &edges)
 {
   int jump,round,k;
   
   int P =1; 
   int dsize=dimensions_.size();

   for (int i=0;i<dsize;i++)
     P*=dimensions_[i];
   if (!direct_)
     for (int i=0;i<P;i++)
       vertices[i][dsize]=nndpersw_;
   
   // here jump denotes the difference between processor numbers  and we decrease  gradually to 1.  
   // e.g. in a 2x3x4x5 torus  jump will start with 120/2=60, andconnect vertices 0--60, 1--61, ..
   // then it will be 60/3=20 and connect 0 --20 etc.     
   jump=round=P;
   for (int i=0;i<dsize;i++){
     jump=jump/dimensions_[i];
     if (wrapped_[i]) {  /* Torus network with edges wrapped around */
       for(int j=0;j<P;j++) {
         k=j-jump;
         if (! vertices[j][i]) /* wrap around edge from 0 to the top  */
           k+=round;
         edges.insert(edge(vertices[j],vertices[k]));
         
         k=j+jump;
         if (vertices[j][i]==dimensions_[i]-1) /* wrap around edge from the top to zero */ 
           k-=round;
         edges.insert(edge(vertices[j],vertices[k]));
       }
     }
     else {  /* Mesh like network with no wrap around edges */
       for(int j=0;j<P;j++) { 
         if ( vertices[j][i]>0)
           edges.insert(edge(vertices[j],vertices[j-jump]));
         if (vertices[j][i] <dimensions_[i]-1)  
           edges.insert(edge(vertices[j],vertices[j+jump]));
       }
     }
     round=jump;
   }
 }
  
// --------------------------------------------------------------------------------------------------------
  // Connects nodes and switches   
  
 void
 hdtorus::compute_node_edges(std::vector <std::vector<int> > &vertices, set<edge> &edges, int nndpersw)
  {
    int dsize=dimensions_.size();

    std::vector<int> vi(dsize+1); // vi specifies a node
    
    for (int i=0; i<n()/nndpersw; i++) {
      for(int j=0;j<dsize;j++)
        vi[j] = vertices[i][j];
      vertices[i][dsize]=nndpersw;
      for (vi[dsize] = 0; vi[dsize] < nndpersw; vi[dsize]++) {
        edges.insert(edge(vi,vertices[i])); // node to switch
        edges.insert(edge(vertices[i],vi)); // switch to node
      }
    }
  }

  // --------------------------------------------------------------------------------------------------------
  /*  Generates a high dimensional torus network
   Inputs:  
   dimensionarray: specifies the number of processors in each dimension 
   and whether  the links are wrapped around  (torus) or not (mesh).  
   dimensionarray[i] <0 refers to absence of  wrap around edges. 
   dimensionarray[i] >0 refers to presence of wrap around edges. 
   
   nndpersw is the number of nodes per switch. 
   widtharrg specifies the width between the switches and the  nodes. 
  */
hdtorus::hdtorus(vector<int> dimensionarray,int nndpersw, int widtharg)
  {
    int P,i;
   
    // Pre-process  the input; 
    int dsize=dimensionarray.size();
    std::vector <std::vector<int> > vertices;

    dimensions_.resize(dsize);
    wrapped_.resize(dsize);
    
    for(i=0;i<dsize;i++) 
      if (dimensionarray[i]<0) {
        wrapped_[i]= 0;
        dimensions_[i]=-dimensionarray[i];
      }
      else if (dimensionarray[i] >2){
        dimensions_[i]=dimensionarray[i]; 
        wrapped_[i]=1;
      }
      else if (dimensionarray[i]==2){
        dimensions_[i]=dimensionarray[i]; 
        wrapped_[i]=0;
      }
      else 
        std::cerr << "Weird dimension " << dimensionarray[i] << "\n"; 

    direct_ = (nndpersw <= 0);
    nndpersw_ = nndpersw;
    width = widtharg;
    
    // Generate the  nodes. 
    for(i=0,P=1;i<dsize;P*=dimensions_[i++]);

    vertices.resize(P);
    if (direct_)
       for(i=0;i<P;vertices[i++].resize(dsize));
    else
       for(i=0;i<P;vertices[i++].resize(dsize+1));
    
    compute_vertex_numbers(vertices, P);
    
    if (direct_) { // each node is connected to the  network directly. 
      compute_vertices(vertices, nodes_);
      set<edge> edges;
      compute_switch_edges(vertices, edges);
      g_.init(nodes_, edges);
    }
    else {  // each node is connected to the network  through a switch
      for (int i=0; i<nndpersw; i++) 
        compute_vertices(vertices, nodes_, i);
      
      std::set<vertex> all(nodes_);
      // insert the switch vertices
      compute_vertices(vertices, all, nndpersw);
      
      set<edge> edges;
      compute_switch_edges(vertices, edges);
      
      compute_node_edges(vertices, edges,nndpersw);
      
      g_.init(all, edges);
    }
    
    init();
    if (!direct_) 
      init_node_to_switch();
  } 
  
// --------------------------------------------------------------------------------------------------------
  
  /// Returns the diameter of the interconnect.
  int hdtorus::diameter() const {
    int D,i;
    int dsize=dimensions_.size();

    for (D=i=0;i<dsize;i++)
      if (wrapped_[i])
        D+=(dimensions_[i]+1)/2;
      else 
        D+=dimensions_[i];      
    return(D);
  }
  
// --------------------------------------------------------------------------------------------------------

  /// Returns the average distance to all nodes.  
  double hdtorus::a() const {
    std::cout << "Average distance in HD trous is not implemented yet "; 
    return 0.0; 
  }


// --------------------------------------------------------------------------------------------------------

  /// Returns the minimum number of edges sliced by bisection.

  int hdtorus::b() const {
    int x;
    int dsize=dimensions_.size();
    int mymin=dimensions_[0];

    if (wrapped_[0])
      mymin*=2;
    for (int i=1; i<dsize;i++){
      x=dimensions_[i];
      if (wrapped_[i])
        x*=2;
      if (mymin <x)
        mymin=x;
    }
    return (mymin);
  } 
// --------------------------------------------------------------------------------------------------------
// The following is an internal function that is used when computing a route between two vertices. 
// It computes the difference  along with the direction to travel the shortest distance while 
// taking into account the wrapped edges.  <0 means route by decreasing the  processors number 
// (e.g., go north) and >0 means  increase the processor number (e.g., go south)

  void hdtorus::compute_diffvect(const vertex &v1, const vertex &v2, vector<int> &diffvect) const
{
  int d1;
  int dsize=dimensions_.size();
 
  for (int i=0;i<dsize;i++) {
    d1=v2.v()[i]-v1.v()[i];
     
    if (d1<0) {
      if (wrapped_[i] && dimensions_[i]< -2*d1)
        diffvect[i]=dimensions_[i]+d1;
      else 
        diffvect[i]=d1;
    }
    else if (d1>0){
      if (wrapped_[i] && dimensions_[i]< 2*d1)
        diffvect[i]=-dimensions_[i]+d1;
      else 
        diffvect[i]=d1;
    }
    else 
      diffvect[i]=0;
  } 
}
  
  // -------------------------------------------------------------------------------------------------------- 
 
  int hdtorus::shortest_path_length(const vertex &v1, const vertex &v2, int id ) const  
  {
    int D,i;
    int dsize=dimensions_.size();

    vector <int> diffvect(dsize);
    if (! id) 
      compute_diffvect(v1,v2, diffvect);
   
    for (D=i=0;i<dsize;i++)
      D+=abs(diffvect[i]);
    return(D);
  }
  
 // --------------------------------------------------------------------------------------------------------
  /** Fills in the walk for the route between two node vertices.  The
      default implementation uses shortest path routing. */
  // virtual void route(const vertex &v1, const vertex &v2, walk &w) const;
 
  void hdtorus::shortest_path(const vertex&v1, const vertex&v2, walk&w) const
  {
    int i,j;
    int dsize=dimensions_.size();
    bool print_route=0;
    vector<vertex_index> vertices;
    vector<int> v(v1.v().size()); 
    vector <int> diffvect(dsize);
    vector<lwedge> nxt;
    
    compute_diffvect(v1,v2,diffvect); 
    
    if (v1 == v2) {
      w.init(vertices);
      return;
    }
    if (print_route) {
      printf(" The path is ");    
      for (i=0;i<v1.v().size();i++)
        printf("%d ",v1.v()[i]);
    }
    
    vertices.push_back(g().vertex_to_index(vertex(v1)));
    for (i=0;i<v.size();++i) 
      v[i]=v1.v()[i];
    
    if (!direct_) {
      // if this is a node, we must first get to a switch
      if (v[dsize] < nndpersw_) {
        v[dsize]= nndpersw_;
         if (print_route) {
           printf(" ~~> ");
           for (i=0;i<v.size();i++)
             printf("%d ",v[i]);
         }
         vertices.push_back(g().vertex_to_index(vertex(v)));
         
      }
      else 
        std::cerr << " Something wrong:message starts at a switch \n" ;
    }
    
    for ( i=0;i<dsize;i++)       {
      if (diffvect[i] <0)
        for(j=diffvect[i];j<0;j++) {
          if (v[i])
            v[i]--;
          else
            v[i]=dimensions_[i]-1;
          if (print_route) {
            printf("--> ");
            for (int k=0;k<v.size();k++)
              printf("%d ",v[k]); 
          }
          vertices.push_back(g().vertex_to_index(v)); 
          
        }
      else 
        for(j=0; j<diffvect[i]; j++) {
          v[i]++;
          if (v[i]==dimensions_[i])
            v[i]=0;
           if (print_route) {
             printf("--> ");
             for (int k=0;k<v.size();k++)
               printf("%d ",v[k]);
           }
           
          vertices.push_back(g().vertex_to_index(v)); 
         
        }
    }
    if (!direct_ ) {
       if (print_route) {
         printf("~~> ");
         for (i=0;i<v2.v().size();i++)
           printf("%d ",v2.v()[i]);
       }
       vertices.push_back(g().vertex_to_index(vertex(v2)));
    }
    w.init(vertices);
     if (print_route) 
       printf("\n");
  }
   
  
  /** Returns all edges that take us one hop from the current
        location and closer to the destination.
        @param loc The vertex for the current location.
        @param dst The destination vertex.
        @param nxt A vector of hops.
        The default implementation throws.
  */
void hdtorus::next_hop(vertex_index loc, vertex_index dst, std::vector<lwedge> &nxt) const
{ 
  int i,t;
  int dsize=dimensions_.size();
  nxt.clear();
  
  if (loc == dst) {
    return;
  }
  
  const vertex &locv = g().index_to_vertex(loc);
  const vertex &dstv = g().index_to_vertex(dst);
  
  vertex hop(g().word_length());
  for (int i=0;i<dsize; ++i)
     hop.v()[i] = locv.v()[i];
  
  // If this is not direct, handle the special cases of hops to or
  // from a node.
  if (!direct_) {
    if (locv.v()[dsize] < nndpersw_) {
      // loc is a node; next hop is a switch
      
      hop.v()[dsize] = nndpersw_;
      nxt.push_back(g().generate_lwedge(loc,hop)); 
      return;
    }
    for (i=0; (i<dsize) && (locv.v()[i] == dstv.v()[i]); i++);
    if (i == dsize) {
      // ready to hop from a switch to the destination node
      nxt.push_back(g().generate_lwedge(loc,dst));
      return;
    }
    hop[dsize] = nndpersw_;
  }
  // direct routing  or not a  speciel case of indirect routing
  for (i=0;i<dsize;i++) { 
    bool flag=true;
    int a = locv.v()[i];
    int b = dstv.v()[i];
    int diff = b - a;
    if (diff) {
      int nx= dimensions_[i];
      if (wrapped_[i]) {
        if (nx== 2*diff  || nx ==-2*diff) {
          // Either way gets us closer in this case.
          flag=false;
          t=hop[i];
          hop[i] = (a + 1)% nx;
          nxt.push_back(g().generate_lwedge(loc,hop));

          hop[i] = (a + nx - 1)% nx;
          nxt.push_back(g().generate_lwedge(loc,hop));
          hop[i]=t;
        }
        
        if      (diff > 0 && 2*diff >  nx) diff = diff-nx;
        else if (diff < 0 && 2*diff < -nx) diff = diff+nx;
      }
      
      if (flag) {
      // this routing applies to both mesh and torus routing
      t=hop[i];
      if (diff>0) hop[i] = (a + 1)%nx;
      else        hop[i] = (a + nx - 1)%nx;
      
      nxt.push_back(g().generate_lwedge(loc,hop));
      hop[i]=t;
      }
    }
  }
}

  // ----------------------------------------------------------------------------------
  
  /// Returns a descriptive name of the graph.
  string hdtorus::name() const
  {
    int dsize=dimensions_.size();
    std::ostringstream ostr;
    ostr << "hdtorus(" ;
    for (int i=0;i<dsize;i++)
      ostr << dimensions_[i] << "," ;
    ostr << ")";
    return ostr.str();
  }

hdtorus::~hdtorus() {
  
} 

boost::shared_ptr<hdtorus> hdtorus::construct(vector<int> dimensions, int nndpersw, int widtharg)
{
 
  return boost::shared_ptr<hdtorus>(new hdtorus(dimensions,nndpersw, widtharg));
}
  

boost::shared_ptr<hdtorus> hdtorus::construct_2D(int n1, int n2, int nndpersw, int widtharg) 
{
  vector<int> dimensions(2);

  dimensions[0]=n1;
  dimensions[1]=n2;
 
  return boost::shared_ptr<hdtorus>(new hdtorus(dimensions,
                                                nndpersw, widtharg));
}
  
boost::shared_ptr<hdtorus> hdtorus::construct_3D(int n1, int n2, int n3, int nndpersw, int widtharg) 
{
  vector<int> dimensions(3);
 
  dimensions[0]=n1;
  dimensions[1]=n2;
  dimensions[2]=n3;
 
  return boost::shared_ptr<hdtorus>(new hdtorus(dimensions,
                                                nndpersw, widtharg));
}
  
boost::shared_ptr<hdtorus> hdtorus::construct_4D(int n1, int n2, int n3, int n4, int nndpersw, int widtharg) 
{
  vector<int> dimensions(4);
 
  dimensions[0]=n1;
  dimensions[1]=n2;
  dimensions[2]=n3;
  dimensions[3]=n4;
  
  return boost::shared_ptr<hdtorus>(new hdtorus(dimensions,
                                                nndpersw, widtharg));
}
  
  
} // end namespace
 



  




