/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */


// transfer_time.h: Base class and specializations for transfer time models.
//
// Author: Curtis Janssen <cljanss@sandia.gov>

#ifndef SSTMAC_NETWORK_TRANSFER_TIME_H
#define SSTMAC_NETWORK_TRANSFER_TIME_H

#include <stdint.h>
#include <iostream>
#include <boost/shared_ptr.hpp>

namespace sstmac {

  class transfer_time {
  public:
    virtual ~transfer_time();
    virtual double time(int64_t nbyte) = 0;
    virtual void print(std::ostream &o=std::cout) = 0;
  };

  class latency_rbandwidth_time: public transfer_time {
  protected:
    latency_rbandwidth_time() {}
  public:
    double latency;
    double rbandwidth;

    static boost::shared_ptr<transfer_time>
      construct(double latency, double rbandwidth);

    /// Returns latency + nbyte * rbandwidth.
    double time(int64_t nbyte);
    void print(std::ostream &o=std::cout);
  };

  class constant_time: public transfer_time {
  protected:
    constant_time() {}
  public:
    double constant;

    static boost::shared_ptr<transfer_time>
      construct(double constant);

    /// Returns constant
    double time(int64_t nbyte);
    void print(std::ostream &o=std::cout);
  };

  class rbandwidth_time: public transfer_time {
  protected:
    rbandwidth_time() {}
  public:
    double rbandwidth;

    static boost::shared_ptr<transfer_time>
      construct(double rbandwidth);

    /// Returns rbandwidth * nbyte
    double time(int64_t nbyte);
    void print(std::ostream &o=std::cout);
  };

  class no_time: public transfer_time {
  protected:
    no_time() {}
  public:
    static boost::shared_ptr<transfer_time>
      construct();
    // This always returns zero.
    double time(int64_t nbyte);
    void print(std::ostream &o=std::cout);
  };

  std::ostream& operator<<(std::ostream&s,const boost::shared_ptr<transfer_time>&);

}

#endif
