/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */


// vertex.h: Interface for vertices in a graph.
//
// Author: Curtis Janssen <cljanss@ca.sandia.gov>

#ifndef SSTMAC_NETWORK_VERTEX_H
#define SSTMAC_NETWORK_VERTEX_H

#include "sstmac/sstmacconfig.h"
#include <iostream>
#include <vector>

#ifdef HAVE_LIBZ
  extern "C" {
    #include <zlib.h>
  }
#endif

namespace sstmac {

  /** This represents the vertices in a graph.
      Each vertex in has a unique label which is given by a vector of
      integers. The number of integers in the label is depends on the
      particular graph. */
  class vertex {
    std::vector<int> v_;
  public:
    vertex() {}
    vertex(int n): v_(n) {}
    vertex(const std::vector<int> &v): v_(v) {}
    void init(const std::vector<int> &v) { v_ = v; }
    const std::vector<int> &v() const { return v_; }
    std::vector<int> &v() { return v_; }
    int &operator [] (int i) { return v_[i]; }
    const int &operator [] (int i) const { return v_[i]; }
    void print(std::ostream &o = std::cout) const;
    bool operator < (const vertex &v) const { return v_ < v.v_; }
    bool operator == (const vertex &v) const { return v_ == v.v_; }
    bool operator != (const vertex &v) const { return v_ != v.v_; }

#ifndef SWIG // Do not parse nested classes with swig.
#ifdef HAVE_LIBZ
    class vertex_hash {
    public:
      std::size_t operator() (const vertex vtx) const {
        uLong crc = crc32(0L, Z_NULL, 0);
        for(unsigned int i=0; i<vtx.v_.size(); ++i) {
          crc = crc32(crc, (Bytef*) &(vtx.v_[i]), sizeof(int));
        }
        return crc;
      }
    };
#endif
#endif // SWIG

  };

  std::ostream& operator<<(std::ostream &o, const vertex &v);

  class vertex_index {
  public:
    typedef int vertex_index_type;
  private:
    vertex_index_type vi_;
  public:
    vertex_index(): vi_(0) {}
    explicit vertex_index(const vertex_index_type&v): vi_(v) {}
    vertex_index(const vertex_index&v): vi_(v.vi_) {}
    //vertex_index_type operator vertex_index_type() const { return vi_; }
    vertex_index operator = (const vertex_index&v) {
      vi_ = v.vi_;
      return *this;
    }
    vertex_index operator = (const vertex_index_type&v) {
      vi_ = v;
      return *this;
    }
    bool operator != (const vertex_index&v) const { return vi_ != v.vi_; }
    bool operator == (const vertex_index&v) const { return vi_ == v.vi_; }
    bool operator < (const vertex_index&v) const { return vi_ < v.vi_; }
    bool operator > (const vertex_index&v) const { return vi_ > v.vi_; }
    bool operator <= (const vertex_index&v) const { return vi_ <= v.vi_; }
    bool operator >= (const vertex_index&v) const { return vi_ >= v.vi_; }
    vertex_index_type value() const { return vi_; }
  };

  inline std::ostream& operator<<(std::ostream&s,const vertex_index&v) {
    s << v.value();
    return s;
  }

}

#endif
