/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_PROCESS_ACTORMBOX_H
#define SSTMAC_PROCESS_ACTORMBOX_H

#include <sstmac/process/simplembox.h>
#include <sstmac/process/actorpid.h>
#include <sstmac/simulator/timestamp.h>
#include <utility>

namespace sstmac {

  // Forward declaration of important types.
  class actortask;
  class actormsg;
  class actorpattern;
  class actorexit;

  /**
   * The base interface for message delivery in an actor-based application.
   * Mailboxes are used in distributed process networks
   * and similar fault-tolerant (or oblivious) communication systems.
   *
   * Mail delivery has a few simple rules:
   * <ul>
   *   <li>Sends:</li>
   *   <ul>
   *     <li>Are non-blocking.</li>
   *     <li>Will succeed if the receiving pid is still alive.</li>
   *     <li>Will quietly fail if the receiving pid is not alive.</li>
   *   </ul>
   *   <li>Receives:<li>
   *   <ul>
   *     <li>Are blocking but have an optional timeout parameter</li>
   *     <li>Are not guaranteed to match the order of sends (if order is
   *         important, you should provide your own index or order parameter</li>
   *     <li>Are matched based on a 'pattern' specified by the receiver.</li>
   *   <li>Links:</li>
   *   <ul>
   *     <li>Create a bi-directional relationship between two processes.</li>
   *     <li>If either process in a linked relationship exits with an
   *         abnormal reason, the exit signal is sent to the other process</li>
   *     <li>A process that does not trap exit signals is killed and all its
   *         linked processes are sent exit signals as well.</li>
   *   </ul>
   *   <li>Monitors:</li>
   *   <ul>
   *     <li>Create a uni-directional relationship with the current process
   *         monitoring the specified process.</li>
   *     <li>If the monitored process exits for any reason, a actordownmsg (object)
   *         is sent to the monitoring process.</li>
   *   </ul>
   * </ul>
   */
  class actormbox : public simplembox {
  public:
    /// So long, and thanks for all the fish.
    virtual ~actormbox() {}

    /// What's my pid.
    virtual actorpid mypid() const = 0;

    /// What nodes are known to the current system?
    virtual const std::vector<nodeid>& nodes() const = 0;

    /// Spawn a process on the given node.  Returns the pid of the new process.
    /// Spawning consists of cloning the input object, assigning it a new pid,
    /// and calling initialize to get the object going.
    /// \param task  The object (thread) that will be spawned.
    /// \param size  How many bytes should we consider this transfer to be?
    /// \return      the pid for the new object, or actorpid::invalid
    ///              if the spawn fails.
    virtual actorpid spawn(const boost::shared_ptr<const actortask>& task,
                         int64_t size) = 0;

    /// Spawn a process on the given node and link it to the current process.
    virtual actorpid spawn_link(const boost::shared_ptr<const actortask>& task,
                              int64_t size) = 0;

    /// Spawn a process on the given node and set up a monitor for it.
    virtual actorpid spawn_monitor(const boost::shared_ptr<const actortask>& task,
                                 int64_t size) = 0;

    /// Create a link to the given target.
    /// This is quietly ignored if the target is no longer in existence.
    /// A link is unique -- an attempt to re-do an existing link is ignored.
    virtual void link(const actorid &target) = 0;

    /// Remove our link to the given target.
    /// Quietly ignored if no such link exists.
    virtual void unlink(const actorid &target) = 0;

    /// Create a monitor for the given target.
    /// This is quietly ignored if the target is no longer in existence.
    /// A monitor is unique -- you will get exactly one actordownmsg when the
    /// process exits (this differs from the behavior of Erlang).
    virtual void monitor(const actorid &target) = 0;

    /// Remove our monitor for the given target.
    /// Quietly ignored if no such monitor exists.
    virtual void unmonitor(const actorid &target) = 0;
    
    /// Send a message to the given pid.
    /// This call always succeeds except for invalid pid.
    /// \throw valueerror if the pid is not valid.
    /// \throw nullerror if sending a NULL message (reserved for recv timeout).
    virtual void send(const actorpid &target, const boost::shared_ptr<const actormsg>& msg,
                      int64_t size) = 0;

    /// Test whether there is any message that matches the given pattern.
    /// This has the same effect as calling recv with a zero-length timeout.
    /// \param source  Our desired source (may be actorpid::any).
    /// \param pat     The pattern we are matching (may be actorpattern::any).
    virtual bool test(const actorpid &source,
                      const boost::shared_ptr<const actorpattern>& pat) const = 0;
    
    /// Block until we get a message matching the given pattern.
    /// \param source  Our desired source (may be actorpid::any).
    /// \param pat     The pattern we are matching (may be actorpattern::any).
    /// \param timeout How long we are willing to wait.  Negative values
    ///                indicate that we will never time out.
    /// \return        The received object (sorry about the dynamic casts)
    ///                or std::make_pair(actorpid::invalid, NULL) if timed out.
    virtual std::pair< actorid, boost::shared_ptr<actormsg> >
    recv(const actorpid &source,
         const boost::shared_ptr<const actorpattern>& pat,
         timestamp timeout = timestamp(-1)) = 0;

    /// Indicate that the current process is exiting for the given reason.
    /// If the reason is not actorexit::normal, all monitors will be sent an
    /// exit message and all linked processes will get a copy of the
    /// exit reason.
    virtual void exit(const boost::shared_ptr<const actorexit>& reason) = 0;
  };

} // end of namespace sstmac

#endif // ! SSTMAC_PROCESS_ACTORMBOX_H
