/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_PROCESS_ACTORSERVER_CLIENT_LINKHANDLER_H
#define SSTMAC_PROCESS_ACTORSERVER_CLIENT_LINKHANDLER_H

#include <sstmac/process/actorserver-client.h>
#include <sstmac/simulator/eventhandler.h>

namespace sstmac {

  /**
   * A private event handler and state manager for making and breaking links.
   */
  class actorserver::client::linkhandler : public eventhandler {
    /// Are we making or breaking a link.
    /// This is mainly here for debug reasons -- the actions are the same.
    bool maker_;
    /// Keep track of whether this request has been satisfied.
    int done_;
    /// We use a scheduler to context switch a caller in and out.
    boost::shared_ptr<scheduler> sched_;
    /// And we use a key to tell the scheduler which thread to operate on.
    boost::shared_ptr<key> key_;

    /// Construct a new handler.
    linkhandler(bool maker, const boost::shared_ptr<scheduler>& sched) :
      maker_(maker), done_(0), sched_(sched), key_(key::construct())
    {}

  public:
    /// Provide a reference counted pointer to a handler to set up a connection.
    static boost::shared_ptr<linkhandler>
    construct_link(const boost::shared_ptr<scheduler>& sched) {
      return boost::shared_ptr<linkhandler>(new linkhandler(true, sched));
    }

    /// Provide a reference counted pointer to a handler to remove a connection.
    static boost::shared_ptr<linkhandler>
    construct_unlink(const boost::shared_ptr<scheduler>& sched) {
      return boost::shared_ptr<linkhandler>(new linkhandler(false, sched));
    }

    /// Wait for this request to complete (context switch).
    void wait() {
      sched_->block(key_);
    }

    /// Indicate that a request has completed successfully.
    /// Note that we may also have already timed out.
    void response_received() {
      if(done_ == 0) {
        done_ = 1;
        sched_->unblock(key_);
      }
    }

    /// Indicate that we have timed out.
    /// Note that we may have already completed the request.
    virtual void handle(const timestamp&) {
      if(done_ == 0) {
        done_ = -1;
        sched_->unblock(key_);
      }
    }

    /// Test whether we completed successfully.
    bool completed() const {
      return done_ > 0;
    }
  };

} // end of namespace sstmac

#endif // ! SSTMAC_PROCESS_ACTORSERVER_CLIENT_LINKHANDLER_H
