/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_PROCESS_ACTORSERVER_H
#define SSTMAC_PROCESS_ACTORSERVER_H

#include <sstmac/process/server.h>
#include <boost/weak_ptr.hpp>
#include <vector>
#include <map>
#include <stdint.h>

namespace sstmac {

  /**
   * Simple server that handles "mailbox" deliveries (nonblocking sends,
   * blocking receives).  Mailboxes are used in distributed process networks
   * and similar fault-tolerant (or oblivious) communication systems.
   * 
   * This is the shared back-end that handles incoming connections to a node,
   * messages for each process (thread, whatever) on this node get forwarded
   * to its 'mbox' object.
   */
  class actorserver : public server {
    /// The actual concrete implementation of a actor mailbox is here.
    /// This allows the internal methods of an actorserver to be
    /// hidden from external types (I'm paranoid).
    class client;
    /// The networkdata type we use to package up our sends.
    class data;
    /// Private message type indicating a request for a spawn call.
    class spawnrequest;
    /// Private message type indicating the completion of a spawn call.
    class spawncomplete;

    /// This object -- needed to allow our clients to have a pointer back.
    boost::weak_ptr<actorserver> self_;

    /// The nodes given to us (each node stored only once).
    std::vector<nodeid> nodes_;

    /// The node we are running on.
    boost::shared_ptr<nodeinterface> node_;

    /// The network we communicate over.
    boost::shared_ptr<networkinterface> net_;
    
    /// The client processes we have (weak pointers, because clients are
    /// welcome to go out of scope any time they want to).
    typedef std::map<actorid, boost::weak_ptr<client> > clientmap_t;
    clientmap_t clients_;

  protected:
    /// Construct a new mailbox server.
    actorserver(const boost::shared_ptr<tasklist>& tlist);

    /* The methods used by our nested client type */
    
  public:
    /// Construct a new mailbox server (one per node).
    /// The nodes in the tasklist are in our initial 'family' of nodes,
    /// more nodes can be added at runtime.
    boost::shared_ptr<actorserver> construct(const boost::shared_ptr<tasklist>& tlist);

    /// Goodbye.
    virtual ~actorserver();

    /* Interface required by the 'server' base class. */

    /// Define a network interface for this server.
    /// This method is called when node::set_server is called with this object.
    virtual void initialize(const boost::shared_ptr<nodeinterface>& my_node,
                            const boost::shared_ptr<networkinterface>& my_net);

    /// Complete a receive.
    /// Forwarded by the node when an incoming message arrives.
    /// It is the responsibility of the server to figure out how to
    /// understand any envelope information in the networkdata.
    virtual void incoming_message(const timestamp &start_time,
                                  const nodeid &sender,
                                  const boost::shared_ptr<networkdata>& load);

  };

} // end of namespace sstmac

#endif // ! SSTMAC_PROCESS_ACTORSERVER_H
