/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/process/cond.h>
#include <sstmac/process/scheduler.h>
#include <sstmac/process/key.h>
#include <sstmac/simulator/timestamp.h>
#include <sstmac/simulator/eventhandler.h>
#include <sstmac/simulator/eventmanager.h>
#include <sstmac/util/errors.h>
#include <algorithm>
#include <iostream>
#include <assert.h>

namespace sstmac {

  // Pointer type.
  using boost::shared_ptr;
  using boost::weak_ptr;

  //
  // Nested type to handle timeouts.
  //
  class cond::unblocker : public eventhandler {
    /// The parent timeout object.
    shared_ptr<cond> parent_;
    /// The key we are timing out.
    shared_ptr<key> key_;

  protected:
    /// Hi.
    unblocker(const shared_ptr<cond>& parent_in, const shared_ptr<key>& key_in) :
      parent_(parent_in), key_(key_in)
    {}

  public:
    /// Hello.
    static shared_ptr<unblocker>
    construct(const shared_ptr<cond>& parent_in, const shared_ptr<key>& key_in) {
      return shared_ptr<unblocker>(new unblocker(parent_in, key_in));
    }

    /// Goodbye.
    virtual ~unblocker() throw() {}

    /// We're done.
    void handle(const timestamp&) {
      parent_->timeout(key_);
    }
  };

  //
  // Private callback to indicate that the given keys has timed out.
  //
  void cond::timeout(const shared_ptr<key> &the_key) {
    keylist_t::iterator it = std::find(keys_.begin(), keys_.end(), the_key);
    if(it != keys_.end()) {
      keys_.erase(it);
      timed_out_.insert(the_key);
      sched_.lock()->unblock(the_key);
    }
  }

  //
  // Hello.
  //
  cond::cond(const shared_ptr<scheduler>& sched) :
    sched_(sched)
  {}

  //
  // Who am I?
  //
  void cond::set_self(const shared_ptr<cond>& self) {
    if(self.get() != this) {
      throw valueerror("cond::set_self:  'self' pointer "
                       "does not point to 'this' object.");
    }
    self_ = self;
  }

  //
  // This is who I am.
  shared_ptr<cond> cond::get_self() const {
    return self_.lock();
  }

  //
  // Wheee!
  //
  shared_ptr<cond> cond::construct(const shared_ptr<scheduler>& schd) {
    shared_ptr<cond> retval(new cond(schd));
    retval->set_self(retval);
    return retval;
  }

  //
  // It is an error to go out of scope with pending keys
  //
  cond::~cond() throw() {
    if(! keys_.empty()) {
      std::cerr << "cond::~cond:  Going out of scope with threads pending.";
      assert(false);
    }
  }

  //
  // Alert the oldest thread waiting on this request.
  //
  void cond::signal() {
    if(! keys_.empty()) {
      shared_ptr<key> the_key = keys_.front();
      keys_.pop_front();
      sched_.lock()->unblock(the_key);
    }
  }

  //
  // Alert all threads waiting on this request.
  //
  void cond::broadcast() {
    if(! keys_.empty()) {
      // Make a copy of the current keys so that newly awoken threads can
      // safely wait on this condition variable.
      keylist_t bcast_to;
      bcast_to.swap(keys_);
      keylist_t::iterator it, end = bcast_to.end();
      for(it = bcast_to.begin(); it != end; ++it) {
        sched_.lock()->unblock(*it);
      }
    }
  }

  //
  // Request that the current context yield until it gets signal()-ed
  //
  void cond::wait() {
    shared_ptr<key> the_key = key::construct();
    keys_.push_back(the_key);
    sched_.lock()->block(the_key);
  }

  //
  // Request that the current context yield unless timed out.
  //
  bool cond::timed_wait(const timestamp &maxwait,
                        const shared_ptr<eventmanager>& mgr)
  {
    shared_ptr<key> the_key = key::construct();
    keys_.push_back(the_key);
    shared_ptr<eventhandler> timer =
      unblocker::construct(this->get_self(), the_key);
    mgr->schedule(mgr->now() + maxwait, timer);
    sched_.lock()->block(the_key);
    // Return true if this is a normal wake-up, false if we timed out.
    bool retval = true;
    if(timed_out_.find(the_key) != timed_out_.end()) {
      timed_out_.erase(the_key);
      retval = false;
    }
    return retval;
  }

} // end of namespace sstmac.
