/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_PROCESS_COND_H
#define SSTMAC_PROCESS_COND_H

#include <sstmac/util/nocopy.h>
#include <boost/shared_ptr.hpp>
#include <boost/weak_ptr.hpp>
#include <deque>
#include <set>

namespace sstmac {

  // Important types.
  class timestamp;
  class scheduler;
  class key;
  class eventmanager;

  /**
   * A condition variable for the lightweight thread class.
   * Requests to wait() on the condition variable result in the calling
   * process being blocked using scheduler::block (only works on 
   * lightweight contexts, not on main (system) context).
   */
  class cond : public nocopy {
    /// Private nested type to handle a timed_wait.
    class unblocker;

    /// This object.
    boost::weak_ptr<cond> self_;

    /// A condition variable is associated with exactly one scheduler.
    /// This is a weak pointer since the scheduler has a strong pointer
    /// to the thread object and the thread has a strong pointer
    /// to the cond object.
    boost::weak_ptr<scheduler> sched_;

    /// The keys associated with this lock.
    /// Each pending thread gets assigned a new key.
    /// A request to signal() wakes the oldest entry in the list (first in),
    /// a request to broadcast() wakes all entries in the list.
    typedef std::deque< boost::shared_ptr<key> > keylist_t;
    keylist_t keys_;

    /// Keep track of threads that have timed out.
    std::set< boost::shared_ptr<key> > timed_out_;

    /// Private callback method used by the nested unblocker type to
    /// indicate that the given keys has timed out.
    void timeout(const boost::shared_ptr<key> &the_key);

  protected:
    /// Hello.
    cond(const boost::shared_ptr<scheduler>& sched);

    /// Who am I?
    void set_self(const boost::shared_ptr<cond>& self);

    /// This is who I am.
    boost::shared_ptr<cond> get_self() const;

  public:
    /// Wheee!
    static boost::shared_ptr<cond> construct(const boost::shared_ptr<scheduler>& schd);

    /// It is an error to go out of scope with pending keys
    /// (since threads should have used a reference-counted pointer
    /// to grab the lock, been context-switched out, and still be in scope).
    virtual ~cond() throw();

    /// Alert the oldest thread waiting on this request.
    /// Has no effect if no threads are waiting.
    void signal();

    /// Alert all threads waiting on this request.
    /// Has no effect of no threads are waiting.
    void broadcast();

    /// Request that the current context yield until it gets signal()-ed
    /// or broadcast()-ed.
    void wait();

    /// Request that the current context yield until it gets signal()-ed
    /// or the given timestamp arrives.
    /// \return true if the thread woke up normally, false if it timed out.
    bool timed_wait(const timestamp &maxwait,
                    const boost::shared_ptr<eventmanager>& mgr);

    /// Return the number of threads currently waiting on this condition.
    int waiting() const {
      return keys_.size();
    }
  };

} // end of namespace sstmac

#endif // ! SSTMAC_PROCESS_COND_H
