/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_PROCESS_LOCK_H
#define SSTMAC_PROCESS_LOCK_H

#include <sstmac/util/nocopy.h>
#include <boost/shared_ptr.hpp>
#include <boost/weak_ptr.hpp>
#include <deque>

namespace sstmac {

  // Important types.
  class scheduler;
  class key;
  class eventmanager;
  class timestamp;

  /**
   * A thread lock for the lightweight class thread.
   * If the lock is engaged, the current thread of execution gets suspended
   * using scheduler::block (can only block lightweight contexts, not main
   * (system) context).  Main context can safely use the
   * try_grab() and release() methods.
   *
   * Pending threads get unlocked in first-in-first-out order
   */
  class lock : public nocopy {
  public:
#ifndef SWIG // Do not parse nested classes with swig.
    /// Make grab and release for the given lock follow program scope
    /// (release the lock automatically).
    /// Use 'lock::automatic scoped(the_lock);' to scope locks (I know; ugly!).
    class automatic : public nocopy {
      boost::shared_ptr<lock> lock_;
    public:
      /// Grab the lock.
      explicit automatic(const boost::shared_ptr<lock>& the_lock,
                         bool already_locked = false) :
        lock_(the_lock)
      {
        if(! already_locked)
          lock_->grab();
      }
      virtual ~automatic() throw() {
        lock_->release();
      }
    };
#endif

  private:
    /// Private nested type to handle a timed_grab.
    class unlocker;

    /// A lock is associated with exactly one scheduler.
    /// This is a weak pointer since the scheduler has a strong pointer
    /// to the thread object and the thread has a strong pointer
    /// to the lock object.
    boost::weak_ptr<scheduler> sched_;

    /// State flag.  Set to true when the lock is engaged.
    bool locked_;

    /// The keys associated with this lock.
    /// Each pending thread gets assigned a new key.
    std::deque< boost::shared_ptr<key> > keys_;

  protected:
    /// Hello.
    lock(const boost::shared_ptr<scheduler>& sched);

  public:
    /// Wheee!
    static boost::shared_ptr<lock> construct(const boost::shared_ptr<scheduler>& schd);

    /// It is an error to go out of scope with pending threads
    /// (since threads should have used a reference-counted pointer
    /// to grab the lock, been context-switched out, and still be in scope).
    virtual ~lock() throw();

    /// Engage the lock (unfortunately, this method cannot have the same
    /// name as the class type).
    /// \throw illformederror if this method is not called from a ucontext
    /// thread or if the scheduler is not valid (out of scope).
    void grab();

    /// Try to grab the lock.
    /// \return true if the lock was grabbed, false if it is currently locked.
    bool try_grab();

    /// Grab the lock unless the given event comes up first.
    /// This one requires the same lock to be given to both the scheduler
    /// and a timed event (the scheduler ignores unblock(...) requests
    /// not associated with a thread).
    bool timed_grab(const timestamp &maxwait,
                    const boost::shared_ptr<eventmanager>& mgr);

    /// Release the lock.
    /// Just like pthread_mutex_unlock, this method does not have to be
    /// called from within the thread that grabbed the lock.
    void release();
  };

} // end of namespace sstmac.

#endif // ! SSTMAC_PROCESS_LOCK_H
