/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/process/mpibtreescatter.h>
#include <sstmac/process/mpiqueue.h>
#include <sstmac/process/mpistatus.h>
#include <sstmac/process/mpicollpayload.h>
#include <sstmac/simulator/eventhandler.h>
#include <sstmac/simulator/nodeinterface.h>
#include <sstmac/util/btscatteriterator.h>
#include <sstmac/util/errors.h>
#include <sstmac/util/debug.h>
#include <boost/lexical_cast.hpp>

namespace sstmac {

  // Pointer types.
  using boost::shared_ptr;
  using boost::weak_ptr;
  using boost::dynamic_pointer_cast;
  using boost::lexical_cast;

  //
  // Given an absolute rank, return the relative rank.
  //
  int mpibtreescatter::relrank(const mpiid &therank) const {
    return ((therank.id - root_.id + comm_.size().id) % comm_.size().id);
  }

  //
  // Given a relative rank, return the absolute rank.
  //
  mpiid mpibtreescatter::absrank(int therank) const {
    return mpiid((therank + root_.id) % comm_.size().id);
  }

  //
  // Start our zero or one receive..
  //
  void mpibtreescatter::init_recv(const timestamp &now) {
    if(debugger::active())
      debug() << "mpibtreescatter[" << this << "]::init_recv(" << now << ")\n";
    // Figure out where we are receiving from.
    const int relative_rank = relrank(comm_.rank());
    if(relative_rank == 0) {
      // We are root -- charge straight to start_send.
      this->init_sends(now);
      return;
    }
    int keymask = 1;
    while(! (relative_rank & keymask)) {
      keymask <<= 1;
    }
    const int relative_source = relative_rank - keymask;
    int endrecv = relative_rank + keymask;
    if(endrecv > comm_.size().id)
      endrecv = comm_.size().id;
    const int entries = endrecv - relative_rank;
    const int length = entries * recvcount_;
    ++pending_recvs_;
    this->start_recv(length, recvtype_, tag_, absrank(relative_source));
  }

  //
  // Start our zero or more sends.
  //
  void mpibtreescatter::init_sends(const timestamp &now) {
    if(debugger::active())
      debug() << "mpibtreescatter[" << this << "]::init_sends(" << now << ")\n";
    // We should never get here with a pending receive.
    if(pending_recvs_)
      throw illformederror("mpibtreescatter::init_sends:  Pending receive.");
    const int relative_rank = relrank(comm_.rank());
    const int size = comm_.size().id;
    int keymask = 1;
    if(relative_rank == 0) {
      // Root node -- we are interested in one-past MSB
      int v = size;
      int shift = 0;
      while(v) {
        ++shift;
        v >>= 1;
      }
      keymask = 1 << shift;
    }
    else {
      while(! (relative_rank & keymask)) {
        keymask <<= 1;
      }
      if(debugger::active())
        debug() << " -- node " << comm_.rank() << " with relative rank "
              << relative_rank << " has keymask " << keymask << "\n";
    }
    // Figure out initial send/receive pair (if any).
    int localmask = keymask >> 1;
    int endsend = relative_rank + keymask;
    if(endsend > size) {
      endsend = size;
    }
    // And start.
    while(localmask > 0) {
      int relative_dest = relative_rank + localmask;
      if(relative_dest >= size) {
        localmask >>= 1;
        continue;
      }
      const int entries = endsend - relative_dest;
      const int length = entries * sendcount_;
      ++pending_sends_;
      mpiid rrank = absrank(relative_dest);
      this->start_send(length, sendtype_, tag_, rrank, content_);
      endsend = relative_dest;
      localmask >>= 1;
    }
    // Some of the nodes are not doing any sends.
    if(pending_sends_ == 0 && pending_recvs_ == 0)
      mpicollective::complete(now, content_->get(comm_.rank()));
  }

  //
  // Hi.
  //
  mpibtreescatter::
  mpibtreescatter(const shared_ptr<key>& thekey,
                  const shared_ptr<mpiqueue>& queue,
                  int sendcount, const mpitype &sendtype,
                  int recvcount, const mpitype &recvtype,
                  const mpiid &root, const mpitag &tag,
                  const mpicomm &comm,
                  const shared_ptr<const mpicollpayload>& content) :
    mpicollective(thekey, queue, comm),
    sendcount_(sendcount), sendtype_(sendtype),
    recvcount_(recvcount), recvtype_(recvtype),
    root_(root), tag_(tag), comm_(comm),
    content_(mpicollpayload::construct(comm.size().id)),
    pending_sends_(0), pending_recvs_(0)
  {
    static const std::string here("mpibtreescatter");
    if(comm.rank() == root) {
      if(! content)
        throw nullerror(here + ": NULL content pointer at root.");
      const mpicollpayload::content_t &cont = content->value();
      mpicollpayload::content_t::const_iterator it, end = cont.end();
      mpiid rr(0);
      for(it = cont.begin(); it != end; ++it, ++rr) {
        shared_ptr<const payload> pload;
        if(*it) pload = (*it)->clone();
        content_->set(rr, pload);
      }
    }
  }
  
  //
  // Callback method to indicate that a send operation has completed.
  //
  void mpibtreescatter::send_complete(const timestamp &now,
                                     int count,
                                     const mpitype &type,
                                     const mpitag &tag,
                                     const mpiid &dest)
  {
    if(debugger::active())
      debug() << "mpibtreescatter[" << this << "]::send_complete(" << now<< ", "
              << count << ", " << type << ", " << tag << ", " << dest << ")\n";
    --pending_sends_;
    if(! pending_sends_) {
      // All done.
      if(debugger::active())
        debug() << " -- sends complete on node " << comm_.rank() << "\n";
      //std::cerr << "Node " << comm_.rank() << ":  After send, content size"
      //        << "is " content_->value().size() << "\n";
      mpicollective::complete(now, content_->get(comm_.rank()));
    }
  }

  //
  // Callback method to indicate that a receive operation has completed.
  //
  void mpibtreescatter::recv_complete(const timestamp &now,
                                      int count,
                                      const shared_ptr<const mpistatus>& stat)
  {
    if(debugger::active())
      debug() << "mpibtreescatter[" << this << "]::recv_complete(" << now<< ", "
              << count << ", " << stat << ")\n";
    shared_ptr<const mpicollpayload> other =
      dynamic_pointer_cast<const mpicollpayload>(stat->content());
    const mpicollpayload::content_t &inval = other->value();
    for(size_t i = 0; i < inval.size(); ++i) {
      if(inval.at(i)) content_->set(mpiid(i), inval.at(i));
    }
    this->start_immediate_copy(count, stat->recv_type(),
                               stat->tag(), stat->source());
  }

  //
  // Completion of a copy event.
  //
  void mpibtreescatter::copy_complete(const timestamp &now,
                                      int count, const mpitype &type,
                                      const mpitag &tag, const mpiid &source)
  {
    if(debugger::active())
      debug() << "mpibtreescatter[" << this << "]::copy_done(" << now << ", "
              << count << ", " << type << ", " << tag << ", " << source
              << ")\n";
    --pending_recvs_;
    if(! pending_recvs_)
      this->init_sends(now);
  }

  //
  // Hello.
  //
  shared_ptr<mpibtreescatter> mpibtreescatter::
  construct(const shared_ptr<key>& thekey,
            const shared_ptr<mpiqueue>& queue,
            int sendcount, const mpitype &sendtype,
            int recvcount, const mpitype &recvtype,
            const mpiid &root, const mpitag &tag,
            const mpicomm &comm, const shared_ptr<const mpicollpayload>& content)
  {
    shared_ptr<mpibtreescatter> retval
      (new mpibtreescatter(thekey, queue, sendcount, sendtype,
                           recvcount, recvtype, root, tag, comm, content));
    retval->set_self(retval);
    return retval;
  }
  
  //
  // Goodbye.
  //
  mpibtreescatter::~mpibtreescatter() throw() {
  }

  //
  // Start this kernel.
  //
  void mpibtreescatter::start(const timestamp &now,
                              const shared_ptr<nodeinterface>& my_node)
  {
    if(debugger::active())
      debug() << "mpibtreescatter[" << this << "]::start(" << now
              << ", " << my_node << ")\n";
    mpicollective::start(now, my_node);
    this->init_recv(now);
  }


} // end of namespace sstmac
