/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_PROCESS_MPICOLLECTIVE_FUNNELEDCOPY_H
#define SSTMAC_PROCESS_MPICOLLECTIVE_FUNNELEDCOPY_H

#include <sstmac/process/mpicollective.h>
#include <sstmac/simulator/eventhandler.h>
#include <sstmac/simulator/nodeinterface.h>
#include <sstmac/util/errors.h>
#include <sstmac/util/debug.h>
#include <deque>

namespace sstmac {

  /**
   * Utility type to handle funneled (one at a time) copying into
   * a shared data array.  Simulates accumulating data from remote sources.
   */
  class mpicollective::funneledcopy : public eventhandler {
    /// Stash copy arguments for future use.
    struct request {
      int count;
      mpitype type;
      mpitag tag;
      mpiid source;

      request(int c, const mpitype &t, const mpitag &g, const mpiid &s) :
        count(c), type(t), tag(g), source(s)
      {}
    };

    // This object -- needed to submit nodeinterface::copy requests.
    boost::weak_ptr<funneledcopy> self_;
    // The parent keeps us alive -- let's avoid cyclic references
    boost::weak_ptr<mpicollective> parent_;
    // When this object is pending, we want to keep the parent alive.
    boost::shared_ptr<mpicollective> keepalive_;
    // The node interface.
    boost::shared_ptr<nodeinterface> node_;
    // Any pending copy requests.
    std::deque<request> pending_;
    
    // Run the next pending request.
    void submit(const timestamp &now, const request &active) {
      node_->copy(now, active.count * active.type.size, self_.lock());
    }

  protected:
    // Hi.
    funneledcopy(const boost::shared_ptr<mpicollective>& parent,
                 const boost::shared_ptr<nodeinterface>& node):
      parent_(parent), node_(node)
    {}

    // This is who I am.
    void set_self(const boost::shared_ptr<funneledcopy>& self) {
      if(self.get() != this)
        throw valueerror("mpicollective::funneledcopy::set_self:   'self' pointer "
                         "does not point to 'this' object.");
      self_ = self;
    }

  public:
    // Hello.
    static boost::shared_ptr<funneledcopy>
    construct(const boost::shared_ptr<mpicollective>& parent,
              const boost::shared_ptr<nodeinterface>& node)
    {
      boost::shared_ptr<funneledcopy> retval(new funneledcopy(parent, node));
      retval->set_self(retval);
      return retval;
    }
    // Bye.
    virtual ~funneledcopy() throw() {}

    // Add a new copy requests.
    void copy(int count, const mpitype &type,
              const mpitag &tag, const mpiid &source)
    {
      if(pending_.empty()) {
        keepalive_ = parent_.lock();
        if(keepalive_ == 0) {
          // Parent has gone out of scope -- bad bad bad.
          throw nullerror("mpicollective::funneledcopy::copy:  NULL parent.");
        }
      }
      pending_.push_back(request(count, type, tag, source));
      if(pending_.size() == 1) {
        // We need to resubmit.
        timestamp now = node_->now();
        this->submit(now, pending_.front());
      }
    }

    // Event completion.
    void handle(const timestamp &now) {
      keepalive_->copy_complete(now, pending_.front().count,
                                pending_.front().type,
                                pending_.front().tag,
                                pending_.front().source);
      pending_.pop_front();
      if(pending_.empty()) {
        // We just finished the last request -- release the keepalive pointer.
        keepalive_ = boost::shared_ptr<mpicollective>();
      }
      else {
        this->submit(now, pending_.front());
      }
    }
  };

} // end of namespace sstmac

#endif // ! SSTMAC_PROCESS_MPICOLLECTIVE_FUNNELEDCOPY_H
