/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/process/mpicollective.h>
#include <sstmac/process/mpicollective-sender.h>
#include <sstmac/process/mpicollective-recver.h>
#include <sstmac/process/mpicollective-funneledcopy.h>
#include <sstmac/process/mpicollective-immediatecopy.h>
#include <sstmac/process/mpistatus.h>
#include <sstmac/process/mpicollpayload.h>
#include <sstmac/process/mpiqueue.h>
#include <sstmac/util/errors.h>
#include <sstmac/util/debug.h>

namespace sstmac {

  // Pointer types.
  using boost::shared_ptr;
  using boost::weak_ptr;

  //
  // Hi.
  //
  mpicollective::mpicollective(const shared_ptr<key>& the_key,
                               const shared_ptr<mpiqueue>& queue,
                               const mpicomm &comm) :
    mpikernel(the_key), status_(mpistatus::construct()),
    queue_(queue), comm_(comm)
  {}

  //
  // Who am I?
  //
  void mpicollective::set_self(const shared_ptr<mpicollective>& self) {
    if(self.get() != this) {
      throw valueerror("mpicollective::set_self:  'self' pointer "
                       "does not point to 'this' object.");
    }
    self_ = self;
  }

  //
  // This is who I am.
  //
  shared_ptr<mpicollective> mpicollective::get_self() {
    return self_.lock();
  }

  //
  // Request a send.
  //
  void mpicollective::start_send(int count,
                                 const mpitype &type,
                                 const mpitag &tag,
                                 const mpiid &dest,
                                 const shared_ptr<const payload>& content)
  {
    if(debugger::active())
      debug() << "mpicollective[" << this << ", comm=" << comm_ 
              << "]::start_send(" << count << ", " << type << ", " << tag
              << ", " << dest << ", " << content << ")\n";
    shared_ptr<sender> snd =
      sender::construct(get_self(), status_, count, type, tag, dest);
    queue_->send(count, type, dest, tag, comm_,
                 false, false, mpimessage::collective, snd, content);
  }

  //
  // Request a recv.
  //
  void mpicollective::start_recv(int count,
                                 const mpitype &type,
                                 const mpitag &tag,
                                 const mpiid &source)
  {
    if(debugger::active())
      debug() << "mpicollective[" << this << ", comm=" << comm_ 
              << "]::start_recv(" << count
              << ", " << type << ", " << tag << ", " << source << ")\n";
    shared_ptr<recver> rcv =
      recver::construct(get_self(), count, type, tag, source);
    queue_->recv(count, type, source, tag, comm_,
                 mpimessage::collective, rcv);
  }

  //
  // Request a funneled copy.
  //
  void mpicollective::start_funneled_copy(int count,
                                          const mpitype &type,
                                          const mpitag &tag,
                                          const mpiid &partner)
  {
    if(funnel_ == 0) {
      throw nullerror("mpicollective::start_funneled_copy: Null funnel object. "
                      "Did you forget to call mpicollective::start?");
    }
    funnel_->copy(count, type, tag, partner);
  }

  //
  // Request an immediate copy.
  //
  void mpicollective::start_immediate_copy(int count,
                                           const mpitype &type,
                                           const mpitag &tag,
                                           const mpiid &partner)
  {
    shared_ptr<eventhandler> copier =
      immediatecopy::construct(this->get_self(), count, type, tag, partner);
    node_->copy(node_->now(), count*type.size, copier);
  }

  //
  // Goodbye.
  //
  mpicollective::~mpicollective() throw() {
  }

  //
  // Get busy child.
  //
  void mpicollective::start(const timestamp &now,
                            const shared_ptr<nodeinterface>& my_node)
  {
    if(debugger::active())
      debug() << "mpicollective[" << this << "]::start(" << now 
              << ", " << my_node << ")\n";
    if(self_.lock() == 0) {
      throw nullerror("mpicollective::start:  Null self pointer.  "
                      "Did you forget to call set_self()?");
    }
    funnel_ = funneledcopy::construct(this->get_self(), my_node);
    node_ = my_node;
  }

} // end of namespace sstmac
