/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_PROCESS_MPICOMMDATA_H
#define SSTMAC_PROCESS_MPICOMMDATA_H

#include <sstmac/process/mpicommid.h>
#include <sstmac/simulator/taskid.h>
#include <sstmac/simulator/nodeid.h>
#include <sstmac/process/mpiid.h>
#include <sstmac/util/nocopy.h>
#include <boost/shared_ptr.hpp>
#include <boost/weak_ptr.hpp>
#include <vector>

namespace sstmac {

  /**
   * Shared data for a communicator.
   *
   * There is no guarantee that this object will be shared by all instances
   * of a communicator (eventually we will want to move to a distributed
   * simulator), but it is best to ensure that the communicator data
   * is not replicted any more than it needs to be.
   */
  class mpicommdata : virtual public nocopy {
    /// A reference-count replacement for 'this' object pointer.
    boost::weak_ptr<mpicommdata> self_;

    /// The identifier for this communicator.
    mpicommid id_;

    /// The tasks participating in this communicator.
    const std::vector<taskid> peers_;
    
    /// The nodes participating in this communicator.
    const std::vector<nodeid> nodes_;

  protected:
    /// Hello.
    mpicommdata(mpicommid id,
                const std::vector<taskid> &peers,
                const std::vector<nodeid> &nodes);

    /// Who am I?
    void set_self(const boost::shared_ptr<mpicommdata>& self);

    /// This is who I am.
    boost::shared_ptr<mpicommdata> self();

  public:
    /// Build a new object.
    static boost::shared_ptr<mpicommdata>
    construct(const mpicommid &id,
              const std::vector<taskid> &peers,
              const std::vector<nodeid> &nodes);

    /// Goodbye.
    virtual ~mpicommdata();

    /// My identifier.
    const mpicommid& id() const {
      return id_;
    }

    /// The peers in this communicator.
    const std::vector<taskid>& peers() const {
      return peers_;
    }

    /// The nodes each peer is running on.
    const std::vector<nodeid>& nodes() const {
      return nodes_;
    }
  };

} // end of namespace sstmac

#endif // ! SSTMAC_PROCESS_MPICOMMDATA_H
