/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_PROCESS_MPICOMMFACTORY_PENDING_H
#define SSTMAC_PROCESS_MPICOMMFACTORY_PENDING_H

#include <sstmac/process/mpicommfactory.h>
#include <sstmac/simulator/nodeid.h>
#include <sstmac/util/nocopy.h>

namespace sstmac {

  // Relevant types.
  class mpicommdata;
  class key;

  /**
   * Manage information associated with pending requests for construction of
   * MPI communicators.
   */
  class mpicommfactory::pending : virtual public nocopy {
    /// The key associated with this request.
    boost::shared_ptr<key> key_;

    /// The scheduler to whom the MPI tasks belong.
    boost::shared_ptr<scheduler> scheduler_;

  protected:
    /// The identifier for the mpi group we're building.
    mpicommid id_;

    /// The group we are using to build from.
    mpicomm template_;

    /// Constructor.
    pending(const mpicommid &new_id, const boost::shared_ptr<scheduler>&);

    /// Block current caller.
    void block();

    /// Unblock all callers.
    void unblock();

  public:
    /// Goodbye.
    virtual ~pending() {}
  };

  /**
   * Supervise calls corresponding to MPI_Comm_create.
   */
  class mpicommfactory::pending_create : public mpicommfactory::pending {
    /// The tasks (jobs) to be included in this group.
    std::vector<taskid> target_peers_;

    /// The nodes involved in the group.
    std::vector<nodeid> target_nodes_;
    
    /// Mapping from node index in the input group to node indices
    /// in the newly created group.
    /// Set to -1 for all nodes that are not members of the new group.
    std::vector<mpiid> index_mapping_;

    /// The number of processes that have entered the add(...) method.
    size_t added_;

    /// The shared mpicommdata.  Gets constructed when everybody
    /// has called add.
    boost::shared_ptr<mpicommdata> commdata_;

  protected:
    /// Constructor.
    /// \param new_id is the identifier for the new group.
    pending_create(const mpicommid &new_id, const boost::shared_ptr<scheduler>& sched);

  public:
    /// Goodbye.
    virtual ~pending_create() {}

    /// Construct a new object of this type.
    static boost::shared_ptr<pending_create>
    construct(const mpicommid &new_id, const boost::shared_ptr<scheduler>& sched);

    /// Get a new communicator for the target group.
    /// Blocks the caller until the group is complete.
    /// All members in the old group must make this call; nodes that
    /// are not in the new group will get mpicomm::null_comm.
    /// \throw valueerror if the members don't match earlier requests.
    mpicomm build(const mpicomm &caller, const std::vector<mpiid> &members);
  };

  /**
   * Supervise calls corresponding to MPI_Comm_split.
   */
  class mpicommfactory::pending_split : public mpicommfactory::pending {
    /// Nested type to manage the key, taskid, and nodeid for each color.
    class keylist;

    /// This is where the mechanics of aggregating entries by color
    /// and building the communicators takes place.
    std::map< int, boost::shared_ptr<keylist> > colors_;

    /// Nodes still pending.
    int pending_;

    /// The maximum number of communicators we're allowed to generate.
    int maxsize_;

    /// The identifier for the next communicator constructed.
    mpicommid current_id_;

  protected:
    /// Constructor.
    pending_split(const mpicommid &new_id, int maxsize, 
                  const boost::shared_ptr<scheduler>& sched);

  public:
    /// Goodbye.
    virtual ~pending_split() {}
    
    /// Hello.
    static boost::shared_ptr<pending_split>
    construct(const mpicommid &new_id, int maxsize,
              const boost::shared_ptr<scheduler>& sched);

    /// Add a participant.  If the call is not complete yet, the caller
    /// gets context-switched out pending completion of the call.
    mpicomm build(const mpicomm &caller, int color, int key);
  };

} // end of namespace sstmac

#endif // ! SSTMAC_PROCESS_MPICOMMFACTORY_PENDING_H
