/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/process/mpidirectscatterv.h>
#include <sstmac/process/mpicollpayload.h>
#include <sstmac/util/errors.h>
#include <sstmac/util/debug.h>
#include <boost/lexical_cast.hpp>

namespace sstmac {

  // Pointer types.
  using boost::shared_ptr;
  using boost::weak_ptr;
  using boost::dynamic_pointer_cast;
  using boost::lexical_cast;
  
  //
  // Hi.
  //
  mpidirectscatterv::mpidirectscatterv(const shared_ptr<key>& thekey,
                                       const shared_ptr<mpiqueue>& queue,
                                       const std::vector<int> &sendcount,
                                       const mpitype &sendtype,
                                       int recvcount, const mpitype &recvtype,
                                       const mpiid &root, const mpitag &tag,
                                       const mpicomm &comm,
                                       const shared_ptr<const mpicollpayload>& cont) :
    mpicollective(thekey, queue, comm),
    sendcount_(sendcount), sendtype_(sendtype),
    recvcount_(recvcount), recvtype_(recvtype),
    root_(root), tag_(tag), comm_(comm),
    content_(mpicollpayload::construct(comm.size().id)),
    pending_sends_(0), pending_recvs_(0)
  {
    static const std::string here("mpibdirectscatterv");
    if(comm.rank() == root) {
      if(! cont)
        throw nullerror(here + ": NULL content pointer at root.");
      const mpicollpayload::content_t &data = cont->value();
      for(size_t i = 0; i < data.size(); ++i) {
        shared_ptr<const payload> pload = data.at(i);
        if(pload) pload = pload = pload->clone();
        content_->set(mpiid(i), pload);
      }
    }
  }
  
  //
  // Callback method to indicate that a send operation has completed.
  //
  void mpidirectscatterv::send_complete(const timestamp &now, int count,
                                       const mpitype &type, const mpitag &tag,
                                       const mpiid &dest)
  {
    if(debugger::active())
      debug() << "mpidirectscatterv::send_complete(" << now << ", " << count
              << ", " << type << ", " << tag << ", " << dest << ")\n";
    --pending_sends_;
    if(! pending_sends_)
      mpicollective::complete(now, content_->get(comm_.rank()));
  }

  //
  // Callback method to indicate that a receive operation has completed.
  //
  void mpidirectscatterv::recv_complete(const timestamp &now, int count,
                                        const shared_ptr<const mpistatus>& stat)
  {
    if(debugger::active())
      debug() << "mpidirectscatterv::recv_complete(" << now << ", " << count
              << ", " << stat << ")\n";
    shared_ptr<const mpicollpayload> other =
      dynamic_pointer_cast<const mpicollpayload>(stat->content());
    const mpicollpayload::content_t &inval = other->value();
    for(size_t i = 0; i < inval.size(); ++i) {
      if(inval.at(i)) content_->set(mpiid(i), inval.at(i));
    }
    this->start_immediate_copy(count, stat->recv_type(),
                               stat->tag(), stat->source());
  }

  //
  // Callback to signal copy completion (only on clients).
  //
  void mpidirectscatterv::copy_complete(const timestamp &now, int count,
                                        const mpitype &type, const mpitag &tag,
                                        const mpiid &source)
  {
    if(debugger::active())
      debug() << "mpidirectscatterv::copy_done(" << now << ", " << count << ", "
              << type << ", " << tag << ", " << source << ")\n";
    --pending_recvs_;
    if(! pending_recvs_)
      mpicollective::complete(now, content_->get(comm_.rank()));
  }

  //
  // Hello.
  //
  shared_ptr<mpidirectscatterv>
  mpidirectscatterv::construct(const shared_ptr<key>& thekey,
                               const shared_ptr<mpiqueue>& queue,
                               const std::vector<int> &sendcount,
                               const mpitype &sendtype,
                               int recvcount, const mpitype &recvtype,
                               const mpiid &root, const mpitag &tag,
                               const mpicomm &comm,
                               const shared_ptr<const mpicollpayload>& content)
  {
    shared_ptr<mpidirectscatterv> retval
      (new mpidirectscatterv(thekey, queue, sendcount, sendtype,
                             recvcount, recvtype, root, tag, comm, content));
    retval->set_self(retval);
    return retval;
  }

  //
  // Goodbye.
  //
  mpidirectscatterv::~mpidirectscatterv() throw() {
  }

  //
  // Start this kernel.
  //
  void mpidirectscatterv::start(const timestamp &now,
                                const shared_ptr<nodeinterface>& my_node)
  {
    if(debugger::active())
      debug() << "mpidirectscatterv::start(" << now << ", " << my_node << ")\n";
    mpicollective::start(now, my_node);
    if(comm_.rank() == root_) {
      // Root node -- fire off a bunch of sends.
      mpiid size = comm_.size();
      pending_sends_ = comm_.size().id - 1;
      if(pending_sends_ == 0) {
        mpicollective::complete(now, content_->get(comm_.rank()));
      }
      else {
        for(mpiid dest(0); dest < size; ++dest) {
          if(dest != root_) {
            this->start_send(sendcount_.at(dest.id), sendtype_, tag_, dest,
                             content_);
          }
        }
      }
    }
    else {
      // Client node -- fire off a single receive.
      pending_recvs_ = 1;
      this->start_recv(recvcount_, recvtype_, tag_, root_);
    }
  }

} // end of namespace sstmac
