/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/process/mpilaunch.h>
#include <sstmac/process/mpistrategy.h>
#include <sstmac/process/mpicommfactory.h>
#include <sstmac/process/mpimagic.h>
#include <sstmac/processor/computeapi.h>
#include <sstmac/simulator/tasklist.h>
#include <sstmac/process/mpiserver.h>
#include <sstmac/process/mpiqueue.h>
#include <sstmac/simulator/nodeinterface.h>
#include <sstmac/process/scheduler.h>
#include <sstmac/process/mpiapp.h>
#include <sstmac/process/mpiapi.h>
#include <sstmac/simulator/environment.h>
#include <sstmac/util/errors.h>
#include <boost/lexical_cast.hpp>

namespace sstmac {

  // Pointer types.
  using boost::shared_ptr;
  using boost::weak_ptr;
  using boost::dynamic_pointer_cast;
  using boost::lexical_cast;

  //
  // Define the object that you want to run.
  //
  mpilaunch::mpilaunch(const shared_ptr<mpiapp>& peertype, bool use_compute_api,
                       int mpirun_np) :
    strategy_(mpistrategy::construct()),
    factory_(mpicommfactory::construct()),
    peer_template_(peertype),
    do_compute_(use_compute_api),
    mpirun_np_(mpirun_np)
  {
    if(peer_template_ == 0) {
      throw valueerror("mpilaunch():  Null peer template type.");
    }
  }

  //
  // Build me an object.
  //
  shared_ptr<mpilaunch> mpilaunch::construct(const shared_ptr<mpiapp>& peertype,
                                             bool use_compute_api,
                                             int mpirun_np)
  {
    shared_ptr<mpilaunch> retval
      (new mpilaunch(peertype, use_compute_api, mpirun_np));
    retval->set_self(retval);
    return retval;
  }

  //
  // Goodbye.
  //
  mpilaunch::~mpilaunch() throw() {
  }

  //
  // What are we?
  //
  const std::string& mpilaunch::description() const {
    static const std::string desc("mpilaunch");
    return desc;
  }

  //
  // Access (with modification rights) the mpi message strategy.
  //
  shared_ptr<mpistrategy> mpilaunch::get_mpistrategy() {
    if( strategy_ == 0 )
      throw nullerror("mpilaunch:get_mpistrategy: null pointer.");
    return strategy_;
  }

  //
  // Define the mpi message strategy.
  //
  void mpilaunch::set_mpistrategy(const shared_ptr<mpistrategy>& strat) {
    if(strat == 0) {
      throw valueerror("mpilaunch::set_mpistrategy: Null strategy pointer.");
    }
    strategy_ = strat;
  }

  //
  // Access the communicator factory.
  //
  shared_ptr<mpicommfactory> mpilaunch::get_mpicommfactory() {
    if( factory_ == 0 )
      throw nullerror("mpilaunch:get_mpicommfactory: null pointer.");   
    return factory_;
  }

  //
  // Define the communicator factory type.
  //
  void mpilaunch::set_mpicommfactory(const shared_ptr<mpicommfactory>& fac) {
    if(fac == 0) {
      throw valueerror("mpilaunch::set_mpicommfactory:  Null factory pointer.");
    }
    factory_ = fac;
  }

  //
  // Fire this object off.
  //
  void mpilaunch::run(const timestamp & /*current_time*/,
                      const shared_ptr<environment>& my_world)
  {
    
    // setup applaunch base class
    run_setup(my_world);
    shared_ptr<scheduler> sched(get_scheduler());

    shared_ptr<mpimagic> the_magic = mpimagic::construct();
    factory_->init(sched, my_world, mpirun_np_);
    shared_ptr<tasklist> tlist = my_world->get_tasklist();
    taskid tasks = tlist->size();
    if(mpirun_np_ > 0) {
      if(mpirun_np_ > tasks.id)
        throw valueerror("mpilaunch::run:  Mpilaunch set up (-np) for "
                         + lexical_cast<std::string>(mpirun_np_)
                         + " nodes, but only "
                         + lexical_cast<std::string>(tasks.id)
                         + " nodes provided at launch.");
      tasks.id = mpirun_np_;
    }
    for(taskid id(0); id < tasks; ++id) {
      shared_ptr<nodeinterface> the_node = my_world->get(id);
      shared_ptr<mpiserver> the_server;
      if(the_node->has_server()) {
        the_server = dynamic_pointer_cast<mpiserver>(the_node->get_server());
        if(the_server == 0) {
          throw illformederror("mpilaunch::run:  Node has a server for "
                               "my ppid, but the server is not mpiserver.");
        }
      }
      else {
        the_server = mpiserver::construct(tlist);
        the_node->set_server(the_server);
      }
      if(the_server == 0) {
        throw illformederror("mpilaunch:  Failed to generate server.");
      }
      shared_ptr<mpiqueue> the_queue = mpiqueue::construct(the_server,the_node);
      the_server->add_task(id, the_queue);
      shared_ptr<mpiapp> the_peer = peer_template_->clone();
      if( do_compute_ ) {
        shared_ptr<computeapi> com_api =
          computeapi::construct(the_peer,the_node);
        the_peer->init_computeapi(com_api);
      }
      shared_ptr<mpiapi> the_api =
        mpiapi::construct(the_peer, the_node, the_queue, the_magic,
                          mpiid(id.id), strategy_, factory_);
      the_peer->init_mpi(the_api);
      sched->add_thread(the_peer);
    }
    // Go.
    sched->run_threads();
  }

} // end of namespace sstmac
