/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_PROCESS_MPIMAGIC_PAYLOADDATA_H
#define SSTMAC_PROCESS_MPIMAGIC_PAYLOADDATA_H

#include <sstmac/process/mpimagic.h>
#include <map>
#include <set>

namespace sstmac {

  /**
   * Internal container to store information about payload provided 
   * during collective operations.
   */
  class mpimagic::payloaddata {
    /// The target size for this payloaddata.
    int target_size_;

    /// The members presently in the container.
    /// During construction, this keeps track of who has joined the container.
    typedef std::map< mpiid, boost::shared_ptr<const payload> > present_t;
    present_t present_;

    /// The ranks that will be retrieving the collected data.
    /// When the 'present' list is full and this list is empty, we are done.
    typedef std::set<mpiid> retrieve_t;
    retrieve_t retrieve_;

    /// When the container is fully populated, this pointer will be
    /// updated to point to the fully formed (shared) result object.
    boost::shared_ptr<mpicollpayload> result_;
    
    /// Initialize a new object.
    payloaddata();

  public:
    /// Get a new reference counted object.
    static boost::shared_ptr<payloaddata> construct();

    /// Clean up.
    virtual ~payloaddata();

    /// Add a new member to the collection.
    void add(mpicomm comm, const boost::shared_ptr<const payload>& data, bool consumer);

    /// Retrieve the collected data for the given member.
    boost::shared_ptr<mpicollpayload> retrieve(mpicomm comm);

    /// Test whether the given members has defined its payload.
    bool rank_populated(mpiid rank) const;

    /// Test whether all members have defined their payload.
    bool fully_populated() const;

    /// Test whether all members have retrieved their data.
    bool fully_retrieved() const;
  };

} // end of namespace sstmac

#endif // ! SSTMAC_PROCESS_MPIMAGIC_PAYLOADDATA_H
