/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/process/mpimagic.h>
#include <sstmac/process/mpimagic-payloaddata.h>
#include <sstmac/process/scheduler.h>
#include <sstmac/process/key.h>
#include <sstmac/process/mpicollpayload.h>
#include <sstmac/util/debug.h>
#include <sstmac/util/errors.h>

namespace sstmac {

  // Pointer types.
  using boost::shared_ptr;

  //
  // Private nested type to track barrier keys.
  //
  mpimagic::barrierkey::barrierkey(mpitag tt, mpicomm cc) :
    ttag(tt), tcomm(cc)
  {}

  //
  // Compare two barrier keys in an associative container.
  //
  bool mpimagic::barrierkey::operator<(const barrierkey &other) const {
    if(ttag != other.ttag)
      return tcomm.id() < other.tcomm.id();
    else
      return ttag < other.ttag;
  }

  //
  // Private nested type to track barrier calls.
  //
  class mpimagic::barrier {
    boost::shared_ptr<key> key_;
    std::vector<bool> joined_;
    int join_cnt_;
    typedef std::set< boost::shared_ptr<scheduler> > sched_t;
    sched_t sched_;
    /// Hi.
    barrier();
  public:
    /// Hello.
    static shared_ptr<barrier> construct() {
      return shared_ptr<barrier>(new barrier);
    }
    /// Add a member.
    /// \return true if the barrier is done.
    bool join(const mpicomm &comm, const boost::shared_ptr<scheduler>& sched);
  };
  
  //
  // A private key type for finding matches for collective ops.
  //
  mpimagic::payloadkey::payloadkey(const std::type_info &inf,
                                   mpitag tt, mpicomm cc) :
    tinfo(&inf), ttag(tt), tcomm(cc)
  {}

  //
  // Match payload keys for sorted associative containers.
  //
  bool mpimagic::payloadkey::operator<(const payloadkey &other) const {
    if(*tinfo != *other.tinfo)  return tinfo->before(*other.tinfo);
    else if(ttag != other.ttag) return ttag < other.ttag;
    else return tcomm.id() < other.tcomm.id();
  }

  //
  // Construct a barrier manager.
  //
  mpimagic::barrier::barrier() :
    key_(key::construct()), join_cnt_(0)
  {}

  //
  // Add a member.
  //
  bool mpimagic::barrier::join(const mpicomm &comm, const shared_ptr<scheduler>& sched)
  {
    // Are we the first one in?
    if(joined_.empty()) {
      joined_.resize(comm.size().id);
      join_cnt_ = 0;
    }
    if(joined_.size() != size_t(comm.size().id)) {
      // Uh-oh.
      throw valueerror("mpimagic::barrier::join:  Communicator size is wrong.");
    }
    int id = comm.rank().id;
    if(joined_.at(id)) {
      // Uh-oh.
      throw valueerror("mpimagic::barrier::join:  Peer has already joined.");
    }
    sched_.insert(sched);
    ++join_cnt_;
    joined_[id] = true;
    // Now figure out what we need to do.
    if(join_cnt_ >= int(joined_.size())) {
      // Last one in.  Release everybody.
      for(sched_t::iterator it = sched_.begin(); it != sched_.end(); ++it) {
        (*it)->unblock(key_);
      }
      return true;
    }
    else {
      sched->block(key_);
      return false;
    }
  }

  //
  // Hello.
  //
  mpimagic::mpimagic() {
  }

  //
  // Hi.
  //
  shared_ptr<mpimagic> mpimagic::construct() {
    return shared_ptr<mpimagic>(new mpimagic);
  }

  //
  // Goodbye.
  //
  mpimagic::~mpimagic() throw() {
  }
  
  //
  // Magic synchronization barrier.
  //
  void mpimagic::barrier(mpitag tag, mpicomm comm,
                         const boost::shared_ptr<scheduler>& sched)
  {
    barrierkey bkey(tag, comm);
    // This will construct the barrier object on first access.
    if(barrier_.find(bkey) == barrier_.end())
      barrier_[bkey] = barrier::construct();
    bool done = barrier_[bkey]->join(comm, sched);
    if(done) {
      barrier_.erase(bkey);
    }
  }

  //
  // Magic storage for payload in collective operations.
  //
  void mpimagic::add_collective_payload(const std::type_info &caller,
                                        mpitag tag, mpicomm comm,
                                        const shared_ptr<const payload>& data,
                                        bool consumer)
  {
    payloadkey pkey(caller, tag, comm);
    if(payload_.find(pkey) == payload_.end())
      payload_[pkey] = payloaddata::construct();
    payload_[pkey]->add(comm, data, consumer);
  }

  //
  // Retrieve magically stored payload.
  //
  shared_ptr<const mpicollpayload>
  mpimagic::get_collective_payload(const std::type_info &caller,
                                   mpitag tag, mpicomm comm)
  {
    //std::cerr << "mpimagic::get_collective_payload(" << caller.name()
    //          << ", " << tag << ", " << comm << "\n";
    payloadkey pkey(caller, tag, comm);
    payload_t::iterator it = payload_.find(pkey);
    if(it == payload_.end())
      throw valueerror("mpimagic::get_collective_payload:  Undefined "
                       "caller/tag/comm combination.");
    shared_ptr<const mpicollpayload> rv = it->second->retrieve(comm);
    if(it->second->fully_retrieved()) {
      //std::cerr << "  -- payload is now fully fetched.  Erasing\n";
      payload_.erase(it);
    }
    return rv;
  }

  //
  // Test whether payload is in place for this node.
  //
  bool mpimagic::have_collective_payload(const std::type_info &caller,
                                         mpitag tag, mpicomm comm) const
  {
    //std::cerr << "mpimagic::have_collective_payload(" << caller.name()
    //            << ", " << tag << ", " << comm << "\n";
    payloadkey pkey(caller, tag, comm);
    bool rv = false;
    payload_t::const_iterator it = payload_.find(pkey);
    if(it != payload_.end())
      rv = it->second->rank_populated(comm.rank());
    return rv;
  }
  
  //
  // Test whether payload is in place for all nodes in the communicator.
  //
  bool mpimagic::have_all_collective_payload(const std::type_info &caller,
                                             mpitag tag, mpicomm comm) const
  {
    //std::cerr << "mpimagic::have_all_collective_payload(" << caller.name()
    //            << ", " << tag << ", " << comm << "\n";
    payloadkey pkey(caller, tag, comm);
    bool rv = false;
    payload_t::const_iterator it = payload_.find(pkey);
    if(it != payload_.end())
      rv = it->second->fully_populated();
    else {
      std::cerr << "mpimagic::have_all_collective_payload:  No match found "
                << "for caller==" << caller.name() << ", " << tag
                << ", " << comm << "\n";
      throw illformederror("mpimagic::have_all_collective_payload:  "
                           "No payload handler defined for this collective.");
    }
    return rv;
  }

} // end of namespace sstmac
