/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_PROCESS_MPIMESSAGE_H
#define SSTMAC_PROCESS_MPIMESSAGE_H

#include <sstmac/simulator/networkdata.h>
#include <sstmac/process/messageid.h>
#include <sstmac/process/mpicomm.h>
#include <sstmac/process/mpiid.h>
#include <sstmac/process/mpitag.h>
#include <sstmac/process/mpitype.h>
#include <sstmac/util/nocopy.h>
#include <boost/shared_ptr.hpp>
#include <boost/weak_ptr.hpp>

namespace sstmac {

  // Important types.
  class mpistatus;
  class payload;

  /**
   * A specialization of networkdata that contains envelope information
   * relevant to MPI messaging.
   */
  class mpimessage : public networkdata, virtual public nocopy {
  public:
    /// The category label used to differentiate user data (which can be
    /// matched using wildcards or [later] probed using probe or iprobe)
    /// from internal communications (e.g. collective operations).
    /// The send_ready and recv_ready categories are intended for setting
    /// up rendezvous operations.  The recv_ack category is indented for
    /// completing synchronous send operations.
    enum category { user, internal, collective, data, ack };

  private:
    int64_t envelope_;
    int64_t mintrans_;
    int count_;
    mpitype type_;
    mpiid source_;
    mpiid dest_;
    mpitag tag_;
    mpicommid commid_;
    int seqnum_;
    messageid msgid_;
    bool ssend_;
    category cat_;
    taskid source_task_;
    taskid dest_task_;
    boost::shared_ptr<const payload> content_;

  protected:
    /// Hello.
    mpimessage(int64_t envelope, int64_t mintrans, 
               int count, const mpitype &type, const mpiid &source,
               const mpiid &dest, const mpitag &tag, const mpicommid &commid,
               int seqnum, const messageid &msgid, bool ssend, category cat,
               const taskid &source_task, const taskid &dest_task,
               const boost::shared_ptr<const payload>& content);

  public:
    /// Construct a new message.
    static boost::shared_ptr<mpimessage>
    construct(int64_t envelope, int64_t mintrans, 
              int count, const mpitype &type, const mpiid &source,
              const mpiid &dest, const mpitag &tag, const mpicommid &commid,
              int seqnum, const messageid &msgid, bool ssend, category cat,
              const taskid &source_task, const taskid &dest_task,
              const boost::shared_ptr<const payload>& content);

    /// Goodbye.
    virtual ~mpimessage() throw();

    /// Parent method to get the complemetary part of this message.
    virtual boost::shared_ptr<networkdata> complement() const;

    /// Access the bytes sent (envelope + (abs(count) * type().size)).
    virtual int64_t bytes() const {
      // Pretend that envelope info is 12 bytes.
      // A handshake request is, therefore, only 12 bytes.
      if(count_ > 0) {
        int64_t sbytes = envelope_ + count_ * int64_t(type_.size);
        return (sbytes > mintrans_ ? sbytes : mintrans_);
      }
      else
        return (envelope_ > mintrans_ ? envelope_ : mintrans_);
    }

    /// The source task.  Required by the parent type.
    virtual const taskid& source_task() const {
      return source_task_;
    }

    /// The destination task.  Required by the parent type.
    virtual const taskid& dest_task() const {
      return dest_task_;
    }

    /// The number of elements sent.
    int count() const {
      return count_;
    }

    /// Access the type label associted with this mesage.
    const mpitype& type() const {
      return type_;
    }

    /// Access the index of the sender.
    const mpiid& source() const {
      return source_;
    }

    /// Access the index of the target.
    const mpiid& dest() const {
      return dest_;
    }

    /// Access the tag associated with this message.
    const mpitag& tag() const {
      return tag_;
    }

    /// Access the id of the communicator that owns this message.
    const mpicommid& commid() const {
      return commid_;
    }

    /// Access the sequence number for this message.
    int seqnum() const {
      return seqnum_;
    }

    /// A (hopfully unique) message id.
    const messageid& id() const { return msgid_; }

    /// Return true if the receiver needs to acknowledge receipt.
    bool ssend() const {
      return ssend_;
    }

    /// Access the category label for this message.
    category cat() const {
      return cat_;
    }

    /// Access the payload associated with this message.
    boost::shared_ptr<const payload> content() const {
      return content_;
    }

    /// Create a new status object for this message.
    boost::shared_ptr<mpistatus> build_status() const;
  };

} // end of namespace sstmac.

#endif // ! SSTMAC_PROCESS_MPIMESSAGE_H
