/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_PROCESS_MPIRECVREQUEST_H
#define SSTMAC_PROCESS_MPIRECVREQUEST_H

#include <sstmac/process/recvrequest.h>
#include <sstmac/process/mpimessage.h>
#include <sstmac/process/mpicommid.h>
#include <sstmac/process/mpiid.h>
#include <sstmac/process/mpitag.h>
#include <sstmac/process/mpitype.h>
#include <sstmac/util/nocopy.h>
#include <boost/weak_ptr.hpp>
#include <stdint.h>

namespace sstmac {

  // Important types.
  class eventhandler;
  class mpimessage;

  /**
   * A receive request posted for MPI calls.
   */
  class mpirecvrequest : public recvrequest, virtual public nocopy {
    /// This object.
    boost::weak_ptr<mpirecvrequest> self_;

    /// The category of messages we are waiting for (must match exactly).
    mpimessage::category category_;
    /// The communicator we are expecting a message from (exact match).
    mpicommid commid_;
    /// The source we are expecing a message from (uses mpiid::matches).
    mpiid source_;
    /// The tag we are expecting (uses mpitag::matches).
    mpitag tag_;
    /// The type we are expecting (not required to match, but issues a warning
    /// in case of a type mis-match).
    mpitype type_;
    /// The number of bytes we are expecting (not required to match, but
    /// issues a warning in case of type mis-match).
    int64_t bytes_;
    /// The event handler that gets notified upon completion.
    boost::shared_ptr<eventhandler> finally_;

    /// At completion, this will be the message we received.
    boost::shared_ptr<const mpimessage> data_;
  protected:
    /// Hello.
    mpirecvrequest(mpimessage::category cat,  const mpicommid &commid,
                   const mpiid &source, const mpitag &tag, 
                   const mpitype &type, int64_t bytes,
                   const boost::shared_ptr<eventhandler>& finally);

    /// Who am I?
    void set_self(const boost::shared_ptr<mpirecvrequest>& self);

    /// This is who I am.
    boost::shared_ptr<mpirecvrequest> get_self();

  public:
    /// Give me an object.
    static boost::shared_ptr<mpirecvrequest>
    construct(mpimessage::category cat,  const mpicommid &commid,
              const mpiid &source, const mpitag &tag, 
              const mpitype &type, int64_t bytes,
              const boost::shared_ptr<eventhandler>& finally);

    /// Goodbye.
    virtual ~mpirecvrequest() throw();

    /// Return true if this request has been fulfilled.
    virtual bool is_complete() const;

    /// Get the data received by this request.
    /// Throw illformederror if the request is not complete.
    virtual boost::shared_ptr<const mpimessage> message();

    /// Method required by the recvrequest parent class.
    /// Test whether this request matches a given message.
    virtual bool matches(const boost::shared_ptr<const networkdata>& message);

    /// Complete a receive.
    virtual void complete(const timestamp &completion_time,
                          const boost::shared_ptr<const networkdata>& message);
  };

} // end of namespace sstmac.

#endif // ! SSTMAC_PROCESS_MPIRECVREQUEST_H
