/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/process/mpiserver.h>
#include <sstmac/process/mpiserver-localtransfer.h>
#include <sstmac/simulator/tasklist.h>
#include <sstmac/simulator/nodeinterface.h>
#include <sstmac/simulator/networkinterface.h>
#include <sstmac/process/mpimessage.h>
#include <sstmac/process/mpiqueue.h>
#include <sstmac/simulator/eventhandler.h>
#include <sstmac/util/errors.h>

namespace sstmac {

  // Pointer types.
  using boost::shared_ptr;
  using boost::weak_ptr;
  using boost::dynamic_pointer_cast;

  //
  // Hello.
  //
  mpiserver::mpiserver(const shared_ptr<tasklist>& tlist) :
    tasklist_(tlist)
  {}

  //
  // Who am I?
  //
  void mpiserver::set_self(const shared_ptr<mpiserver>& self) {
    if(self.get() != this) {
      throw valueerror("mpiserver::set_self:  'self' pointer does not point "
                       "to 'this' object.");
    }
    self_ = self;
  }

  //
  // This is who I am.
  //
  shared_ptr<mpiserver> mpiserver::get_self() {
    return self_.lock();
  }

  //
  // Hello.
  //
  shared_ptr<mpiserver> mpiserver::construct(const shared_ptr<tasklist>& tlist) {
    shared_ptr<mpiserver> retval(new mpiserver(tlist));
    retval->set_self(retval);
    return retval;
  }

  //
  // Goodbye.
  //
  mpiserver::~mpiserver() throw() {
  }

  //
  // Define a task for this server.  Each task can only be defined once
  //
  void mpiserver::add_task(const taskid &tid, const boost::shared_ptr<mpiqueue>& peer)
  {
    if(peer == 0) {
      throw nullerror("mpiserver::add_task:  null peer pointer.");
    }
    queue_t::iterator it = queue_.find(tid);
    if(it != queue_.end()) {
      throw illformederror("mpiserver::add_task:  Attempt to redefine peer.");
    }
    queue_[tid] = peer;
  }

  //
  // Test whether the given task is defined on this node.
  //
  bool mpiserver::has_task(const taskid &tid) const {
    return (queue_.find(tid) != queue_.end());
  }

  //
  // Define a network interface for this server.
  //
  void mpiserver::initialize(const boost::shared_ptr<nodeinterface>& my_node,
                             const boost::shared_ptr<networkinterface>& my_net)
  {
    if(my_node == 0) {
      throw nullerror("mpiserver::initialize:  null node pointer.");
    }
    if(my_net == 0) {
      throw nullerror("mpiserver::initialize:  null network pointer.");
    }
    node_ = my_node;
    net_ = my_net;
  }

  //
  // Send data.
  //
  void mpiserver::send(const boost::shared_ptr<mpimessage>& payload,
                       const boost::shared_ptr<eventhandler>& completion)
  {
    taskid recver = payload->dest_task();
    queue_t::iterator it = queue_.find(recver);    
    if(it != queue_.end()) {
      // Transfer between processes on local node.
      // FIXME:  Make this a node_->copy(...) operation (right now too fast).
      node_->copy(node_->now(), payload->bytes(),
                  localtransfer::construct(it->second, payload, completion));
      // it->second->incoming_message(node_->now(), payload);
      // completion->complete(node_->now());
    }
    else {
      // Network transfer.
      if(net_ == 0) {
        throw illformederror("mpiserver::send:  null network pointer.");
      }
      net_->send(node_->now(), tasklist_->node(recver), payload, completion);
    }
  }

  //
  // Complete a receive.
  void mpiserver::incoming_message(const timestamp &start_time,
                                   const nodeid &/*sender*/,
                                   const shared_ptr<networkdata>& load)
  {
    shared_ptr<mpimessage> mess = dynamic_pointer_cast<mpimessage>(load);
    if(mess == 0) {
      throw illformederror("mpiserver::incoming_message:  Message is not "
                           "of type mpimessage.");
    }
    taskid recver = mess->dest_task();
    queue_t::iterator it = queue_.find(recver);
    if(it == queue_.end()) {
      std::cerr << "mpiserver[ " << this << "]::incoming_message:  "
                << "Failed to find " << recver << " in the task queues.  "
                << "My tasks are: [ ";
      for(queue_t::iterator a = queue_.begin(); a != queue_.end(); ++a) {
        std::cerr << a->first << ", ";
      }
      std::cerr << "]\n";
      throw illformederror("mpiserver::incoming_message:  Message for task "
                           "not registered with this server.");
    }
    it->second->incoming_message(start_time, mess);
  }



} // end of namespace sstmac
