/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_PROCESS_MPISTATUS_H
#define SSTMAC_PROCESS_MPISTATUS_H

#include <sstmac/process/status.h>
#include <sstmac/process/mpicomm.h>
#include <sstmac/process/mpiid.h>
#include <sstmac/process/mpitag.h>
#include <sstmac/process/mpitype.h>

namespace sstmac {

  // Forward declaration of important types.
  class payload;

  /**
   * The status of an MPI message (sent or received).
   */
  class mpistatus : public status {
    /// The communicator involved in this communication.
    mpicommid commid_;
    /// The source and destination for this communication.
    mpiid source_, dest_;
    /// The tag that was used.
    mpitag tag_;
    /// The datatype that was communicated.
    mpitype send_type_, recv_type_;
    /// The content sent (may be null).
    boost::shared_ptr<const payload> content_;

  protected:
    /// Hello.
    mpistatus();

    /// Copy constructor.
    mpistatus(const mpistatus &m);

  public:
    /// Make me an object.
    static boost::shared_ptr<mpistatus> construct();

    /// Goodbye.
    virtual ~mpistatus() throw();

    /// Get a copy of this status object.
    boost::shared_ptr<mpistatus> clone() const;
    
    /// The communicator that was used.
    const mpicommid& commid() const;

    /// Register the communicator used.
    void set_commid(const mpicommid &c);

    /// Transmissions originated from here:
    const mpiid& source() const;

    /// Define the source.
    void set_source(const mpiid &s);

    /// Transmissions ended up here:
    const mpiid& dest() const;

    /// Define the destination.
    void set_dest(const mpiid &d);

    /// This was the tag used:
    const mpitag& tag() const;

    /// Define the tag.
    void set_tag(const mpitag &t);

    /// This is the type that was sent:
    const mpitype& send_type() const;

    /// Define the type.
    void set_send_type(const mpitype &t);

    /// This is the type that was received.
    const mpitype& recv_type() const;

    /// Define the received type.
    void set_recv_type(const mpitype &t);

    /// This is the content.
    boost::shared_ptr<const payload> content() const;

    /// Set the content.
    void set_content(const boost::shared_ptr<const payload>& cc);
  };
  
} // end of namespace sstmac.

#endif // ! SSTMAC_PROCESS_MPISTATUS_H
