/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_PROCESS_PROCESSREQUEST_H
#define SSTMAC_PROCESS_PROCESSREQUEST_H

#include <sstmac/process/request.h>
#include <set>

namespace sstmac {

  // Important types.
  class scheduler;
  class key;

  /**
   * A specialization of a request that permits a caller to block
   * while waiting for the completion of a request.
   *
   * The same effect can be achieved through the request::add_listener
   * interface, but this is cleaner (a little more readable).
   */
  class processrequest : public request {
    /// The type used to hold any and all schedulers waiting for me.
    typedef std::set< boost::shared_ptr<scheduler> > schedulers_t;

    /// The scheduler(s) that our process(es) are blocked through.
    schedulers_t sched_;
   
    /// The key used to block my process(es).
    boost::shared_ptr<key> key_;

    /// The completion time passed in to complete.
    timestamp final_time_;

  protected:
    /// Hello.
    processrequest(const boost::shared_ptr<key>& thekey);

  public:
    /// Build me a request.
    static boost::shared_ptr<processrequest>
    construct(const boost::shared_ptr<key>& thekey);

    /// Goodbye.
    virtual ~processrequest() throw();

    /// Block the caller thread until this request is complete.
    /// \throw illformederror if the request is complete.
    const timestamp& wait(const boost::shared_ptr<scheduler>& sched);

    /// Grab the completion request so we can unblock any sleeping threads.
    /// Starts by calling request::complete.
    void complete(const timestamp &current_time, const boost::shared_ptr<const status>&);
  };

} // end of namespace sstmac

#endif // ! SSTMAC_PROCESS_PROCESSREQUEST_H
