/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/process/request.h>
#include <sstmac/simulator/eventhandler.h>
#include <sstmac/util/debug.h>
#include <sstmac/util/errors.h>

namespace sstmac {

  //
  // Hello.
  //
  request::request() :
    active_(true)
  {}

  //
  // Who am I?
  //
  void request::set_self(const boost::shared_ptr<request>& self) {
    if(self.get() != this) {
      throw valueerror("request::set_self:  'self' pointer does not point "
                       "to 'this' object.");
    }
    self_ = self;
  }

  //
  // This is who I am.
  //
  boost::shared_ptr<request> request::get_self() {
    return self_.lock();
  }

  //
  // Construct.
  //
  boost::shared_ptr<request> request::construct() {
    boost::shared_ptr<request> retval(new request);
    retval->set_self(retval);
    return retval;
  }

  //
  // Virtual destructor is essential.
  //
  request::~request() throw() {
  }

  //
  // Test whether the request is complete.
  //
  bool request::is_complete() const {
    return (status_ != 0);
  }

  //
  // Test whether this request is active (can be waited on or retrieved).
  //
  bool request::is_active() const {
    return active_;
  }
  
  //
  // Tag this request as inactive.
  //
  void request::set_inactive() {
    active_ = false;
  }

  //
  // Access the status object.
  //
  boost::shared_ptr<const status> request::get_status() {
    if(! this->is_complete()) {
      std::cerr << "request " << this << " is not complete\n";
      throw illformederror("request::get_status:  Request  is not complete.");
    }
    return status_;
  }

  //
  // Register a listener.
  //
  void request::add_listener(const boost::shared_ptr<eventhandler>& lis) {
    if(! this->is_active()) {
      throw illformederror("request::add_listener:  Request is not active.");
    }
    if(this->is_complete()) {
      throw illformederror("request::add_listener:  Request is complete.");
    }
    if(std::find(listeners_.begin(), listeners_.end(), lis) == listeners_.end())
      listeners_.push_back(lis);
  }

  //
  // Un-register from a callback to the given request.
  //
  void request::erase_listener(const boost::shared_ptr<eventhandler>& lis) {
    std::vector< boost::shared_ptr<eventhandler> >::iterator it = 
      std::find(listeners_.begin(), listeners_.end(), lis);
    if(it != listeners_.end())
      listeners_.erase(it);
  }

  //
  // Complete the request.
  //
  void request::complete(const timestamp &current_time,
                         const boost::shared_ptr<const status>& stat)
  {
    if(debugger::active()) debug() << "request[" << this << "]::complete(" << current_time
              << ", " << stat << ").\n";
    if(stat == 0) {
      throw nullerror("request::complete:  null status pointer.");
    }
    status_ = stat;
    // Notify our listeners.
    if(! listeners_.empty()) {
      std::vector< boost::shared_ptr<eventhandler> >::iterator it,
        end = listeners_.end();
      for(it = listeners_.begin(); it != end; ++it) {
        (*it)->handle(current_time);
      }
      listeners_.clear();
    }
  }

} // end of namespace sstmac
