/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/process/scheduler.h>
#include <sstmac/process/thread.h>
#include <sstmac/process/key.h>
#include <sstmac/simulator/eventhandler.h>
#include <sstmac/util/errors.h>
#include <sstmac/util/debug.h>
#include <iostream>
#include <exception>
#include <iostream>

namespace sstmac {

  using boost::shared_ptr;

  //
  // Hello world.
  //
  scheduler::scheduler(size_t stacksize, size_t alloc_unit,
                       bool use_mprot_guard) :
    running_(false), stackalloc_(stacksize, alloc_unit, use_mprot_guard),
    blocked_count_(0)
  {
    if(debugger::active()) debug() << "scheduler(" << stacksize << ", " << alloc_unit << ", "
            << use_mprot_guard << ")\n";
    init_context(caller_context_);
  }

  //
  // Set the self pointer.
  //
  void scheduler::set_self(const shared_ptr<scheduler>& self) {
    if(self.get() != this) {
      throw valueerror("scheduler::set_self:  'self' pointer does "
                       "not point to 'this' object.");
    }
    self_ = self;
  }

  //
  // Get the self pointer.
  //
  shared_ptr<scheduler> scheduler::get_self() {
    return self_.lock();
  }

  //
  // Allocate an object of this type.
  //
  shared_ptr<scheduler> scheduler::construct(size_t stacksize,
                                             size_t alloc_unit,
                                             bool use_mprot_guard)
  {
    shared_ptr<scheduler> retval
      (new scheduler(stacksize, alloc_unit, use_mprot_guard));
    retval->set_self(retval);
    return retval;
  }

  //
  // Goodbye.
  //
  scheduler::~scheduler() throw() {
    destroy_context(caller_context_);
  }

  //
  // Clear all thread data.
  //
  void scheduler::clear() {
    if(running_) {
      throw illformederror("scheduler::clear called on a running scheduler.");
    }
    if(debugger::active()) debug() << "scheduler::clear:  Setting completion to ignoredevent\n";
    completion_ = eventhandler::ignore;
    running_ = false;
    threads_.clear();
    stackalloc_.clear();
    active_.clear();
    current_ = &caller_context_;
    //blocked_.clear();
    if(blocked_count_ != 0) {
      throw illformederror("scheduler::clear():  Cannot clear when "
                           "blocked processes are outstanding.");
    }
    blocked_count_ = 0;
  }

  //
  // Define completion handler.
  //
  void scheduler::set_completion(const shared_ptr<eventhandler>& completion) {
    if(running_) {
      throw illformederror("scheduler::set_completion:  Refusing to set "
                           "completion handler on a running scheduler.");
    }
    completion_ = completion;
  }

  //
  // Report deadlocked threads.
  //
  void scheduler::report_deadlock() const {
    size_t size = threads_.size();
    for(size_t i = 0; i < size; ++i) {
      if(threads_[i]->get_state() != thread::DONE) {
        std::string state;
        switch(threads_[i]->get_state()) {
        case thread::PENDING: state = "PENDING";  break;
        case thread::ACTIVE:  state = "ACTIVE";   break;
        case thread::DONE:    state = "DONE";     break;
        default:      state = "unknown!"; break;
        }
        std::cerr << "Thread " << i << " is reported as " << state << "\n";
      }
    }
  }
  
  //
  // Add a task to the scheduler.
  //
  void scheduler::add_thread(const shared_ptr<thread>& task) {
    task->init(stackalloc_.alloc(), stackalloc_.stacksize(), 
               self_.lock(), NULL);
    threads_.push_back(task);

    // The context of the task that called the run_threads routine.
    // Set all task pointers to fall back to the next task in line
    // and store the "upstream" linked list of relationships.
    if(debugger::active())
      debug() << "scheduler[" << this << "]::run_threads:  "
              << "Populating context list for task " << task << "\n";
    active_.push_back(threads_.back()->context());
  }

  //
  // Run until no thread can run (all threads finished or blocked).
  //
  void scheduler::run_threads() {
    if(this->running_) {
      throw schedulererror("scheduler::run_threads:  Scheduler is "
                           "currently running.");
    }
    if(threads_.size() < 1) {
      if(debugger::active()) debug() << "Warning:  scheduler::run_threads called "
                << "with no tasks in the queue.\n";
      return;  // Nothing to do.
    }
    // OK, we can start.
    this->running_ = true;
    // Now run like crazy until we have no more active tasks.
    while(! active_.empty()) {
      current_ = active_.front();
      active_.pop_front();
      if(debugger::active())
        debug() << "scheduler[" << this << "]::run_threads: "
                << "Switching to context " << current_ << ", leaving "
                << active_.size() << " contexts pending.\n";
      swap_context(caller_context_, *current_);
    }
    this->running_ = false;
    // Finally, figure out whether we are done with all possible threads.
    if(blocked_count_ == 0) {
      completion_->handle(timestamp(-1));
      this->clear();
    }
  }

  //
  // A task is being signalled as complete.
  //
  void scheduler::complete(void *stack)
  {
    if(! running_) {
      throw schedulererror("scheduler::complete:  Scheduler is not "
                           "currently running a thread.");
    }
    stackalloc_.free(stack);
    if(debugger::active())
      debug() << "scheduler[" << this << "]::complete: "
              << "Context " << current_ << ", is complete, now "
              << active_.size() << " contexts pending.\n";
    swap_context(*current_, caller_context_);
    // This thread is not permitted to start again.
    throw illformederror("scheduler::complete:  Thread switched in again "
                         "after calling complete and relinquishing stack.");
  }

  //
  // Block the current task pending a specific message key.
  //
//   void scheduler::block(const shared_ptr<key>& req) {
//     if(! running_) {
//       throw schedulererror("scheduler::block:  Scheduler is not "
//                            "currently running.");
//     }
//     if(req == 0) {
//       throw nullerror("scheduler::block:  Refusing to block "
//                       "pending a null key pointer.");
//     }
//     // OK, we're ready to roll.
//     //blocked_.insert(std::make_pair(req, current_));
//     req->blocking_.push_back(current_);
//     blocked_count_++;
//     if(debugger::active()) debug() << "scheduler[" << this << "]::block:  Blocking " 
//               << current_ << " pending " << req << "\n";
//     swap_context(*current_, caller_context_);
//     if(debugger::active()) debug() << "scheduler[" << this << "]::block:  Context " << current_
//               << " (which was pending " << req << ") is now active\n";
//   }

  //
  // Unblock and run tasks waiting on the given key.
  //
//   void scheduler::unblock(const shared_ptr<key>& req) {
//     if(debugger::active()) debug() << "scheduler[" << this << "]::unblock(" << req << ")\n";
//     key::blocking_t::iterator it, end = req->blocking_.end();
//     for(it = req->blocking_.begin(); it != end; ++it) {
//       active_.push_back(*it);
//     }
//     blocked_count_ -= req->blocking_.size();
//     req->blocking_.clear();
//     if(! running_) {
//       if(debugger::active())
//         debug() << "scheduler[" << this << "]::unblock:  Running threads\n";
//       this->run_threads();
//     }
//     if(debugger::active())
//       debug() << "scheduler[" << this  << "]::unblock:  Finished\n";
//   }

} // end of namespace sstmac.
