/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_PROCESS_SCHEDULER_H
#define SSTMAC_PROCESS_SCHEDULER_H

#include <sstmac/process/context.h>
#include <sstmac/util/stackalloc.h>
#include <sstmac/util/nocopy.h>
#include <sstmac/util/debug.h>
#include <sstmac/process/key.h>
#include <sstmac/util/errors.h>
#include <boost/shared_ptr.hpp>
#include <boost/weak_ptr.hpp>
#include <vector>
#include <deque>
#include <list>
#include <map>
#include <unistd.h>
#include <string.h>
#include <errno.h>

namespace sstmac {

  // Forward declarations of relevant types.
  class eventhandler;
  class stackalloc;
  class thread;
  class key;

  /**
   * A cooperative scheduler which keeps track of a pool of
   * tasks in a varying mix of active, blocked, and finished.
   */
  class scheduler : virtual public nocopy {
  private:
    typedef boost::shared_ptr<thread> taskptr_t;

    /// This object.
    boost::weak_ptr<scheduler> self_;

    /// Signal this event handler when all threads are done.
    boost::shared_ptr<eventhandler> completion_;

    /// Sentinel to check whether we are currently in the run_threads routine
    /// (it is an error to call that routine recursively).
    bool running_;

    /// The caller context (thread that called scheduler::run).  We go back
    /// to this context on every context switch.
    threadcontext_t caller_context_;

    /// This list ensures that the tasks don't get erased while we're working.
    std::vector<taskptr_t> threads_;

    /// This is how we keep our chucks of stack.
    stackalloc stackalloc_;

    /// The list of active contexts.
    /// Threads get removed from this list on calls to block()
    /// and added back at unblock().
    typedef std::deque<threadcontext_t*> active_t;
    active_t active_;
    /// The currently active position in the list.
    threadcontext_t *current_;

    /// Keep track of blocked contexts and what key they're waiting for.
    //typedef std::multimap< boost::weak_ptr<const key>, threadcontext_t* > blockmap_t;
    //blockmap_t blocked_;
    int blocked_count_;

  protected:
    /// Hello world.
    scheduler(size_t stacksize, size_t alloc_unit, bool use_mprot_guard);
    
    /// Set the "this is you" pointer.
    void set_self(const boost::shared_ptr<scheduler>&);

    /// Get the "this is me" pointer.
    boost::shared_ptr<scheduler> get_self();

  public:
    /// Allocate an object of this type.
    /// \param stacksize is the stack size given to each thread.
    /// \param alloc_unit is the chunks of memory requested in each mmap call.
    static boost::shared_ptr<scheduler> construct(size_t stacksize = 1<<16,
                                                  size_t alloc_unit = 1<<24,
                                                  bool use_mprot_guard = true);

    /// Goodbye, cruel world.
    virtual ~scheduler() throw();

    /// Clear all thread state held by this scheduler.
    void clear();

    /// When all threads have completed, signal this event handler.
    void set_completion(const boost::shared_ptr<eventhandler>& completion);

    /// Threads added to the scheduler.
    int active_threads() const { return (active_.size() + running_); }

    /// Output information on deadlocked threads.
    void report_deadlock() const;

    /// Add a task to the scheduler.
    /// Only one scheduler can "own" a task -- the task object
    /// is supposed to let you know if you are messing this up.
    void add_thread(const boost::shared_ptr<thread>& task);

    /// Run until all threads are blocked or finished.
    /// \throw scheduler_error if this method is called again before
    /// it returns.
    void run_threads();

    /// Complete the given task and return its stack to the allocator.
    void complete(void *stack);

    /// Block the current task pending a specific message key.
    /// This forces the task to yield, until "unblock" is called
    /// with the correct key, at which point the key is
    /// considered fulfilled.
    /// \throw scheduler_error if this method is called when none of the
    /// scheduled threads are running (otherwise we might never find our way
    /// back to the current execution context).
    void block(const boost::shared_ptr<key>& req) {
      if(! running_) {
        throw schedulererror("scheduler::block:  Scheduler is not "
                             "currently running.");
      }
      if(req == 0) {
        throw nullerror("scheduler::block:  Refusing to block "
                        "pending a null key pointer.");
      }
      // OK, we're ready to roll.
      //blocked_.insert(std::make_pair(req, current_));
      req->blocking_.push_back(current_);
      blocked_count_++;
      if(debugger::active())
        debug() << "scheduler[" << this << "]::block: "
                << "Blocking context " << current_ << ", leaving "
                << active_.size() << " contexts pending.\n";
      swap_context(*current_, caller_context_);
    }

    /// Unblock any task waiting on the given key.
    void unblock(const boost::shared_ptr<key>& req) {
      key::blocking_t::iterator it, end = req->blocking_.end();
      for(it = req->blocking_.begin(); it != end; ++it) {
        active_.push_back(*it);
      }
      blocked_count_ -= req->blocking_.size();
      req->blocking_.clear();
      if(! running_) {
        this->run_threads();
      }
    }

    /// The currently active context.
    const threadcontext_t* current_context() const {
      return current_;
    }
  };

} // end of namespace sstmac

#endif // ! SSTMAC_PROCESS_SCHEDULER_H
