/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_PROCESS_SERVER_H
#define SSTMAC_PROCESS_SERVER_H

#include <sstmac/simulator/nodeid.h>
#include <sstmac/simulator/timestamp.h>
#include <boost/shared_ptr.hpp>
#include <boost/weak_ptr.hpp>

namespace sstmac {

  // Important types.
  class nodeinterface;
  class networkinterface;
  class eventhandler;
  class networkdata;

  /**
   * An interface for a network event receiver (server) on a node.
   * Simulates the "back end" (shared resources) of a parallel process,
   * permitting policy for message buffering etc. to be defined on a
   * per-job basis.
   *
   * The "front end" details of the server are left to the discretion of the
   * application (e.g. an mpiserver for mpi applications), this interface
   * only specifies the "back end" needed to deliver incoming network 
   * messages.
   */
  class server {
  public:
    /// Goodbye.
    virtual ~server() {}

    /// Define a network interface for this server.
    /// This method is called when node::set_server is called with this object.
    virtual void initialize(const boost::shared_ptr<nodeinterface>& my_node,
                            const boost::shared_ptr<networkinterface>& my_net) = 0;

    /// Complete a receive.
    /// Forwarded by the node when an incoming message arrives.
    /// It is the responsibility of the server to figure out how to
    /// understand any envelope information in the networkdata.
    virtual void incoming_message(const timestamp &start_time,
                                  const nodeid &sender,
                                  const boost::shared_ptr<networkdata>& load)=0;
  };

} // end of namespace sstmac.

#endif // ! SSTMAC_PROCESS_SERVER_H
