/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/process/simplemboxlaunch.h>
#include <sstmac/process/simplemboxapp.h>
#include <sstmac/process/simplemailserver.h>
#include <sstmac/process/actormagic.h>
#include <sstmac/process/scheduler.h>
#include <sstmac/simulator/tasklist.h>
#include <sstmac/simulator/environment.h>
#include <sstmac/simulator/nodeinterface.h>
#include <sstmac/util/errors.h>

namespace sstmac {

  using boost::shared_ptr;
  using boost::dynamic_pointer_cast;

  //
  // Actual object construction.
  //
  simplemboxlaunch::simplemboxlaunch(const shared_ptr<simplemboxapp>& ptype) :
    peer_template_(ptype)
  {
    if(! ptype)
      throw nullerror("simplemboxlaunch:  NULL peer template.");
  }
  
  //
  // Create a new launcher.
  //
  shared_ptr<simplemboxlaunch>
  simplemboxlaunch::construct(const shared_ptr<simplemboxapp>& ptype)
  {
    shared_ptr<simplemboxlaunch> rv(new simplemboxlaunch(ptype));
    rv->set_self(rv);
    return rv;
  }

  //
  // Clean up.
  //
  simplemboxlaunch::~simplemboxlaunch() throw() {
  }

  // 
  // What are we?
  //
  const std::string& simplemboxlaunch::description() const {
    static const std::string me("simplemboxlaunch");
    return me;
  }
   
  // 
  // Fire this object off.
  //
  void simplemboxlaunch::run(const timestamp & /*current_time*/,
                             const shared_ptr<environment>& my_world)
  {
    // Set up the applaunch base.
    this->run_setup(my_world);
    shared_ptr<scheduler> sched = this->get_scheduler();
    shared_ptr<actormagic> magic = actormagic::construct(sched);
    shared_ptr<tasklist> tlist = my_world->get_tasklist();
    // Start setting up the processes.
    // Note that the user may have specified more than one process per node.
    for(taskid id(0); id < tlist->size(); ++id) {
      shared_ptr<nodeinterface> the_node = my_world->get(id);
      shared_ptr<simplemailserver> the_server;
      if(the_node->has_server()) {
        the_server = 
          dynamic_pointer_cast<simplemailserver>(the_node->get_server());
        if(! the_server)
          throw illformederror("simplemboxlaunch:  Node has a server for "
                               "my ppid, but it is not the right type.");
      }
      else {
        the_server =
          simplemailserver::construct(tlist->node(id), magic, tlist, sched);
        the_node->set_server(the_server);
      }
      // Construct the task.
      shared_ptr<simplembox> the_mbox = the_server->new_mbox();
      shared_ptr<simplemboxapp> the_app = peer_template_->clone();
      the_app->init_mbox(the_mbox);
      sched->add_thread(the_app);
    }
    // Go.
    sched->run_threads();
  }

} // end of namespace sstmac
