/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/process/simplempiapp.h>
#include <sstmac/process/mpiapi.h>
#include <sstmac/process/mpistatus.h>
#include <sstmac/process/mpitype.h>
#include <sstmac/process/mpiop.h>
#include <sstmac/process/mpitag.h>
#include <sstmac/processor/eventdata.h>
#include <sstmac/processor/computeapi.h>
#include <sstmac/util/errors.h>
#include <sstmac/util/debug.h>
#include <string.h>
#include <stdio.h>

namespace sstmac {

  // Pointer types.
  using boost::shared_ptr;

  //
  // Protected constructor.
  //
  simplempiapp::simplempiapp() { }

  //
  // Public constructor.
  //
  shared_ptr<simplempiapp> simplempiapp::construct() {
    shared_ptr<simplempiapp> obj(new simplempiapp());
    obj->set_self(obj);
    return obj;
  }

  //
  // Goodbye.
  //
  simplempiapp::~simplempiapp() throw() {
  }

  //
  // Get a copy.
  //
  shared_ptr<mpiapp> simplempiapp::clone() {
    return simplempiapp::construct();
  }

  //
  // Go.
  //
  void simplempiapp::run() {
    if(mpi() == 0) {
      throw nullerror("simplempiapp::run:  mpiapi pointer is null.");
    }
    mpicomm world = mpi()->comm_world();
    timestamp start = mpi()->init();
    mpiid rank = world.rank();
    mpiid size = world.size();
    mpitag tag(0);
    int count(1);

    // create eventdata objects
    boost::shared_ptr<eventdata> instructions = eventdata::construct();
  
    // process first compute kernel
    instructions->set_event("PAPI_FP_INS",1000);
    instructions->set_event("PAPI_LST_INS",5);
    if( compute_api() == 0 )
      throw nullerror("simplempiapp::run:  computeapi pointer is null.");
    compute_api()->compute(instructions);

    // reduce
    mpi()->allreduce(count,mpitype::mpi_double,mpiop::sum,world);

    // process second compute kernel
    instructions->set_event("PAPI_FP_INS",2000);
    instructions->set_event("PAPI_LST_INS",100);
    compute_api()->compute(instructions);

    // reduce
    mpi()->allreduce(count,mpitype::mpi_double,mpiop::sum,world);   

    timestamp done = mpi()->finalize();

    if(rank.id == 0) {
      double runtime = (done-start).sec();
      const size_t bsize = 2000;
      char *buffer = new char[bsize+1];
      buffer[bsize] = '\0';
      snprintf(buffer, bsize, "Peer %d started at %.12f and finished at %.12f "
               "with a total runtime of %g seconds\n", rank.id, start.sec(),
               done.sec(), runtime);
      if(!fwrite(buffer, sizeof(char), strlen(buffer), stderr)) {
        // Failed to write, but we can't do much about that now.
      }
    }
  }

} // end of namespace sstmac.
