/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_PROCESS_THREAD_H
#define SSTMAC_PROCESS_THREAD_H

#include <sstmac/process/context.h>
#include <sstmac/util/nocopy.h>
#include <boost/shared_ptr.hpp>
#include <boost/weak_ptr.hpp>
#include <unistd.h>

namespace sstmac {

  // Forward declarations of relevant types.
  class scheduler;
  class key;

  /**
   * Base class for tasks that can be cooperative scheduled using threadcontext.
   */
  class thread : virtual public nocopy {
  public:
    /// Help resolve deadlock situations.
    enum state { PENDING, ACTIVE, DONE };

  private:
    friend class scheduler;

    /// This object.
    boost::weak_ptr<thread> self_;
    /// The stack given to this thread.
    void *stack_;
    /// The stacksize.
    size_t stacksize_;
    /// Each thread can only run under one scheduler.
    boost::weak_ptr<scheduler> scheduler_;
    /// The magic.
    threadcontext_t context_;

    /// Monitor state for deadlock detection.
    state state_;
    /// This key gets used when other threads ask to wait (join) this thread.
    boost::shared_ptr<key> key_;

  private:
    /// Private method that gets called by the scheduler.
    void init(void *stack, int stacksize,
              const boost::shared_ptr<scheduler>& sched,
              threadcontext_t *yield_to);

    /// Run routine that defines the initial context for this task.
    /// This routine calls the virtual thread::run method.
    static void run_routine(void *threadptr);

  protected:
    /// Initialize private data.
    thread();

    /// This object.
    virtual void set_self(const boost::shared_ptr<thread>&);

    /// Derived types must provide a way to access their reference-counted
    /// pointer.
    virtual boost::shared_ptr<thread> get_self();

  public:
    virtual ~thread();

    /// Get current thread state.
    state get_state() const { return state_; }

    /// Derived types need to override this method.
    virtual void run() = 0;

    /// A convenience request to start a new thread.
    /// The current thread has to be initialized for this to work.
    void start_thread(const boost::shared_ptr<thread>& thr);

    /// A request to join threads (blocks the calling thread until
    /// this thread is done).  It is an error to call this from within
    /// the current thread.
    void join();

    /// Access the scheduler for this thread.
    boost::shared_ptr<scheduler> get_scheduler();

    /// Test whether the current task has been initialized (activated)
    /// by a scheduler.
    bool initialized() const { return (scheduler_.lock() != 0); }

    /// Access the context for this thread (handle with care).
    inline threadcontext_t* context() { return &context_; }
  };

} // end of namespace sstmac

#endif // ! SSTMAC_PROCESS_THREAD_H
