/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_PROCESS_VALUEPAYLOAD_H
#define SSTMAC_PROCESS_VALUEPAYLOAD_H

#include <sstmac/process/payload.h>

namespace sstmac {

  /**
   * Network payload consisting of a value with sensible copy semantics.
   *
   * Don't use this for pointer types, since you will break assumptions
   * about const-ness or protection of underlying data.
   */
  template <typename T>
  class valuepayload : public payload {
    /// The data we keep.
    T data_;

    /// Construction time.
    valuepayload(const T &inval) :
      data_(inval)
    {}

    /// Construction with type conversion.
    template <typename Other>
    valuepayload(const Other &inval) :
      data_(inval)
    {}

    /// Private copy constructor.
    valuepayload(const valuepayload<T> &other) :
      data_(other.data_)
    {}

  public:
    /// Construct a new valuepayload.
    static typename boost::shared_ptr< valuepayload<T> >
    construct(const T &inval) {
      // Results in ICE in gcc-4.0.
      //return typename boost::shared_ptr< valuepayload<T> >
      //  (new valuepayload(inval));
      return boost::shared_ptr< valuepayload<T> >
        (new valuepayload(inval));
    }

    /// Construct a new valuepayload with type conversion.
    template <typename Other>
    static typename boost::shared_ptr< valuepayload<T> >
    construct(const Other &inval) {
      // Results in ICE in gcc-4.0
      //return typename boost::shared_ptr< valuepayload<T> >
      //  (new valuepayload(inval));
      return boost::shared_ptr< valuepayload<T> >
        (new valuepayload(inval));    }

    /// Clone this object.
    boost::shared_ptr<payload> clone() const {
      // Results in ICE in gcc-4.0
      //return typename boost::shared_ptr< valuepayload<T> >
      //  (new valuepayload(*this));
      return boost::shared_ptr< valuepayload<T> >
        (new valuepayload(*this));
    }

    /// Access the underlying data in a non-const context.
    T& data() {
      return data_;
    }
    
    /// Access the underlying data in a const context.
    const T& data() const {
      return data_;
    }
  };

} // end of namespace sstmac

#endif // ! SSTMAC_PROCESS_VALUEPAYLOAD_H
