/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SIMULATOR_CIRCUIT_EDGE_H
#define SSTMAC_SIMULATOR_CIRCUIT_EDGE_H

#include <sstmac/simulator/circuit-transfer.h>
#include <sstmac/util/debug.h>
#include <iostream>
#include <list>

namespace sstmac {

  /**
   * A handler to manage the state of an edge in a circuit network
   * (reserved vs. available).  Requests for this edge are managed in
   * a FIFO queue.
   *
   * Unlike so many other classes in the simulator, this one is not indended
   * to be individuall reference counted (and must be copy-constructable).
   */
  class circuit::edge {
    /// The state of this edge.
    int nchannel_;
    int nreserved_;

    /// The queue of callers waiting to receive this edge.
    /// Objects are removed from this queue when it is their turn.
    std::list< boost::shared_ptr<transfer> > pending_;
    
  public:
    /// Hi.
    edge(int nchannel = 1) :
      nchannel_(nchannel),
      nreserved_(0)
    {}

    /// I just cannot bring myself to use the default (or non-virtual) dtor.
    virtual ~edge() throw() {
      if(nreserved_ > 0) {
        debug() << "WARNING:  ~edge:  Going out of scope while reserved.\n";
      }
    }

    /// Put in a request for this edge.
    void reserve(const boost::shared_ptr<transfer> &caller) {
      if (debugger::active()) debug() << "circuit::edge[" << this << "]::reserve(" << caller << ")\n";
      if(nreserved_ < nchannel_) {
        if (debugger::active()) debug() << "    edge given to caller " << caller << "\n";
        nreserved_++;
        caller->edge_available(this);
      }
      else {
        if (debugger::active()) debug() << "    caller " << caller << " added to request stack\n";
        pending_.push_back(caller);
      }
    }

    /// Release current reservation on this edge.
    void release() {
      if (debugger::active()) debug() << "circuit::edge[" << this << "]::release\n";
      nreserved_--;
      if(! pending_.empty()) {
        boost::shared_ptr<transfer> next = pending_.front();
        pending_.pop_front();
        nreserved_++;
        if (debugger::active()) debug() << "    edge given to caller " << next
                << " " << nreserved_ << " of " << nchannel_
                << "channels in use\n";
        next->edge_available(this);
      }
      else {
        if (debugger::active()) debug() << ":   no pending callers.  Edge marked available\n";
      }
    }

    /// Set the number of channels that are available for this edge.
    void set_nchannel(int nchannel) { nchannel_ = nchannel; }

  };
  
} // end of namespace sstmac

#endif // ! SSTMAC_SIMULATOR_CIRCUIT_EDGE_H
