/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/simulator/circuit-transfer.h>
#include <sstmac/simulator/circuit-transfer-finish.h>
#include <sstmac/simulator/circuit-edge.h>
#include <sstmac/simulator/eventmanager.h>
#include <sstmac/network/transfer_time.h>
#include <sstmac/network/network.h>
#include <sstmac/network/walk.h>
#include <sstmac/util/errors.h>

namespace sstmac {

  // Types.
  using boost::shared_ptr;
  using boost::weak_ptr;
  using sstmac::walk;

  //
  // Hi.
  //
  circuit::transfer::transfer(const shared_ptr<circuit> &parent,
                              int64_t bytes, const sstmac::walk &the_walk,
                              const timestamp &send_time,
                              const timestamp &recv_time,
                              const timestamp &tran_time,
                              const shared_ptr<eventhandler> &sendcomplete,
                              const shared_ptr<eventhandler> &recvcomplete) :
    parent_(parent), bytes_(bytes),
    send_time_(send_time), recv_time_(recv_time), tran_time_(tran_time),
    sendcomplete_(sendcomplete), recvcomplete_(recvcomplete)
  {
    if(debugger::active()) debug() << "transfer[" << this << "] created\n";
    if(parent_ == 0) {
      throw nullerror("circuit::transfer(...):  null parent pointer.");
    }
    pending_edges_ = the_walk.length();
    edges_.reserve(pending_edges_);
    for(int index = 0; index < pending_edges_; ++index) {
      int edgenum
        = parent_->network_->g().edge_to_index(the_walk.e(index)).value();
      edges_.push_back(edgenum);
    }
  }

  //
  // Who am I?
  //
  void circuit::transfer::set_self(const shared_ptr<transfer> &self) {
    if(self.get() != this) {
      throw valueerror("circuit::transfer::set_self:  'self' pointer "
                       "does not point to 'this' object.");
    }
    self_ = self;
  }

  //
  // This is who I am.
  //
  boost::shared_ptr<circuit::transfer> circuit::transfer::get_self() {
    return self_.lock();
  }
  
  //
  /// Hi.
  //
  shared_ptr<circuit::transfer>
  circuit::transfer::construct(const shared_ptr<circuit> &parent, int64_t bytes,
                               const sstmac::walk &the_walk,
                               const timestamp &send_time,
                               const timestamp &recv_time,
                               const timestamp &tran_time,
                               const shared_ptr<eventhandler> &sendcomplete,
                               const shared_ptr<eventhandler> &recvcomplete)
  {
    shared_ptr<transfer> retval(new transfer(parent, bytes, the_walk,
                                             send_time, recv_time, tran_time,
                                             sendcomplete, recvcomplete));
    retval->set_self(retval);
    return retval;
  }

  //
  // Goodbye.
  //
  circuit::transfer::~transfer() throw() {
    if(debugger::active()) debug() << "~transfer[this]: destroyed\n";
    if(pending_edges_) {
      if(debugger::active()) debug() << "WARNING:  circuit::transfer: Exiting with pending edges.\n";
    }
  }

  //
  // Eventhandler completion.
  //
  void circuit::transfer::handle(const timestamp&) {
    for(size_t index = 0; index < edges_.size(); ++index) {
      parent_->edge_.at(edges_[index]).reserve(this->get_self());
    }
  }
 
  /// A callback used by an edge to indicate that it is now available.
  void circuit::transfer::edge_available(edge *) {
    --pending_edges_;
    if(pending_edges_ == 0) {
      const timestamp &now = parent_->mgr_->now();
      timestamp t_send_done = now + send_time_;
      timestamp t_recv_done = now + recv_time_;
      timestamp t_tran_done = now + tran_time_;
      // We have the full circuit.
      parent_->mgr_->schedule(t_send_done, sendcomplete_);
      parent_->mgr_->schedule(t_recv_done, recvcomplete_);
      parent_->mgr_->schedule(t_tran_done, finish::construct(parent_, edges_));
      // Release any reference counted pointers, just to be sure.
      parent_ = shared_ptr<circuit>();
      edges_.clear();
      sendcomplete_ = shared_ptr<eventhandler>();
      recvcomplete_ = sendcomplete_;
    }
  }

} // end of namespace sstmac.
