/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SIMULATOR_CIRCUIT_TRANSFER_H
#define SSTMAC_SIMULATOR_CIRCUIT_TRANSFER_H

#include <sstmac/simulator/circuit.h>
#include <sstmac/simulator/eventhandler.h>

namespace sstmac {

  // Important types.
  class walk;

  /**
   * An individual data transfer request on a circuit network.
   */
  class circuit::transfer : public eventhandler {
    /// Handler to release the edges when transmission is done.
    class finish;

    /// This object.
    boost::weak_ptr<transfer> self_;

    /// The parent circuit.
    boost::shared_ptr<circuit> parent_;

    /// The number of bytes we're sending.
    int64_t bytes_;

    /// The time we are spending on send, recv, and transfer.
    timestamp send_time_, recv_time_, tran_time_;

    /// The edges we need to visit.
    std::vector<int> edges_;

    /// The signal we will be sending when the sender is done.
    boost::shared_ptr<eventhandler> sendcomplete_;

    /// The signal we will be sending when the receiver is done.
    boost::shared_ptr<eventhandler> recvcomplete_;

#ifdef CIRCUIT_WORM
    /// The next edge (local indexing) we need to grab to complete the circuit.
    size_t next_edge_;
#else
    /// The number of edges we need to wait for before progressing.
    int pending_edges_;
#endif

  protected:
    /// The constructor has to operate in two phases (unfortunately).
    transfer(const boost::shared_ptr<circuit> &parent, int64_t bytes,
             const sstmac::walk &the_walk, const timestamp &send_time,
             const timestamp &recv_time, const timestamp &tran_time,
             const boost::shared_ptr<eventhandler> &sendcomplete,
             const boost::shared_ptr<eventhandler> &recvcomplete);

    /// Who am I?
    void set_self(const boost::shared_ptr<transfer> &self);

    /// This is who I am.
    boost::shared_ptr<transfer> get_self();

  public:
    /// Hi.
    static boost::shared_ptr<transfer>
    construct(const boost::shared_ptr<circuit> &parent, int64_t bytes,
              const sstmac::walk &the_walk, const timestamp &send_time,
              const timestamp &recv_time, const timestamp &tran_time,
              const boost::shared_ptr<eventhandler> &sendcomplete,
              const boost::shared_ptr<eventhandler> &recvcomplete);

    /// Goodbye.
    virtual ~transfer() throw();

    /// Eventhandler completion.
    virtual void handle(const timestamp &now);

    /// A callback used by an edge to indicate that it is now available.
    void edge_available(edge *the_edge);
  };

} // end of namespace sstmac

#endif // ! SSTMAC_SIMULATOR_CIRCUIT_TRANSFER_H
