/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SIMULATOR_CIRCUIT_H
#define SSTMAC_SIMULATOR_CIRCUIT_H

#include <sstmac/simulator/interconnect.h>
#include <sstmac/simulator/bandwidth.h>
#include <vector>

namespace sstmac {

  // Forward declaration of important types.
  class latencymodel;
  class eventhandler;
  class network;
  class transfer_time;

  /**
   * Model network traffic by requiring that all network links
   * between the source and destination be simultaneously available.
   * An idealized network performance model is assumed where the
   * total time a message will keep the circuit open is the latency
   * plus the message size divided by the bandwidth.  The network
   * topology and routes are provided by specializations of the
   * network class.
   *
   * This is roughly equivalent to a circuit network with a separate
   * signaling channel.  The signaling channel (managed through circuit::edge)
   * handles requests for a given edge using a FIFO queue; a message is
   * initiated by placing a reservation for all edges connecting a sender
   * and receiver (using circuit::edge::reserve), and upon completion the edge
   * is explicitly released (using circuit::edge::release).
   */
  class circuit : public interconnect {
  public:
    /// Timing model.  This may need to be re-thought a bit.
    enum timingmodel { ORIGINAL, SPLIT };

  private:
    /// Private class to keep track of whether an edge is available or not
    /// and to signal a sender when the entire circuit is available.
    class edge;

    /// Private class to handle the actual transmission of a message;
    /// place the request, wait for confirmation from the edges, 
    /// "transfer data" (set up completion events for sender and receiver),
    /// and release the edges.
    class transfer;

    /// This object.
    boost::weak_ptr<circuit> self_;

    /// The network topology.  This object should eventually take over the
    /// ownership of per-link latencies and bandwidths.
    boost::shared_ptr<sstmac::network> network_;

    /// Reservation queues for each of the edges in our network.
    /// It is the sstmac::network's job to determine whether edges
    /// are shared in both directions (half duplex) or not (full duplex).
    std::vector<edge> edge_;

    /// The event manager.
    boost::shared_ptr<eventmanager> mgr_;

    /// The latency passed to the ctor. This and the bandwidth
    /// are used to construct timing models.
    boost::shared_ptr<latencymodel> latency_;

    /// The bandwidth passed to the ctor. This and the latency
    /// are used to construct timing models.
    bandwidth bandwidth_;

    /// The timing model used.  Defaults to 'original'.
    timingmodel timing_model_;

  protected:
    /// Construct a circuit with a custom latency model.
    circuit(const boost::shared_ptr<latencymodel> &latency, bandwidth bw,
            const boost::shared_ptr<sstmac::network> &net);

    /// Who am I?
    void set_self(const boost::shared_ptr<circuit> &self);

    /// This is who I am.
    boost::shared_ptr<circuit> get_self();

    /// Use the original timing model.
    void original_timing_model(data payload, 
                               timestamp &start_time, timestamp &send_time,
                               timestamp &recv_time, timestamp &tran_time);

    /// Use the split timing model.
    void split_timing_model(data payload, 
                            timestamp &start_time, timestamp &send_time,
                            timestamp &recv_time, timestamp &tran_time);
  public:
    /// New circuit network with a constant latency.
    /// Legacy interface with double-precision value for bandwidth.
    static boost::shared_ptr<circuit>
    construct(const timestamp latency, double bandwidth,
              const boost::shared_ptr<sstmac::network> &net);

    /// New circuit network with a constant latency.
    static boost::shared_ptr<circuit>
    construct(const timestamp latency, bandwidth bw,
              const boost::shared_ptr<sstmac::network> &net);

    /// New circuit network with a caller-defined latency model.
    static boost::shared_ptr<circuit>
    construct(const boost::shared_ptr<latencymodel> &latency, bandwidth bw,
              const boost::shared_ptr<sstmac::network> &net);

    /// Goodbye.
    virtual ~circuit() throw();

    /// Required method from parent.
    virtual int slots() const;

    /// My event manager.
    virtual void set_eventmanager(const boost::shared_ptr<eventmanager> &mgr);

    /// Send the given message at the given time.
    /// Notify the sender node when the send is complete
    /// and notify the receiver node when receive needs to be performed.
    virtual void send(const timestamp &start_time,
                      int64_t bytes,
                      nodeid sender,
                      nodeid recver,
                      const boost::shared_ptr<eventhandler> &sendhandler,
                      const boost::shared_ptr<eventhandler> &recvhandler);

    
    /// Request original timing model.
    void original_timing_model() { timing_model_ = ORIGINAL; }

    /// Request split timing model.
    void split_timing_model() { timing_model_ = SPLIT; }
   
    /// Get current timing model.
    timingmodel timing_model() const { return timing_model_; }
  };

  // Debug -- gcc-4.5 is having some real issues with trying to 
  // use lexical_cast on circuit.
  inline std::ostream& operator<<(std::ostream &os, const circuit&) {
    os << "[class circuit]";
    return os;
  }

} // end of namespace sstmac

#endif // ! SSTMAC_SIMULATOR_CIRCUIT_H
