/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SIMULATOR_CIRCUIT2_H
#define SSTMAC_SIMULATOR_CIRCUIT2_H

#include <sstmac/simulator/interconnect.h>
#include <sstmac/simulator/data.h>
#include <sstmac/simulator/bandwidth.h>
#include <vector>

namespace sstmac {

  // Forward declaration of important types.
  class eventhandler;
  class network;
  class transfer_time;

  /**
   * Model network traffic by requiring that all network links
   * between the source and destination be simultaneously available.
   * An idealized network performance model is assumed where the
   * total time a message will keep the circuit open is the latency
   * plus the message size divided by the bandwidth.  The network
   * topology and routes are provided by specializations of the
   * network class.
   *
   * This is roughly equivalent to a circuit network with a separate
   * signaling channel.  The signaling channel (managed through circuit::edge)
   * handles requests for a given edge using a FIFO queue; a message is
   * initiated by placing a reservation for all edges connecting a sender
   * and receiver (using circuit::edge::reserve), and upon completion the edge
   * is explicitly released (using circuit::edge::release).
   */
  class circuit2 : public interconnect {
    /// Private class to keep track of whether an edge is available or not
    /// and to signal a sender when the entire circuit2 is available.
    class edge;

    /// Private class to handle the actual transmission of a message;
    /// place the request, wait for confirmation from the edges, 
    /// "transfer data" (set up completion events for sender and receiver),
    /// and release the edges.
    class transfer;

    /// This object.
    boost::weak_ptr<circuit2> self_;

    /// Timing model for time needed to start the send.
    boost::shared_ptr<sstmac::transfer_time> start_time_;

    /// Timing model for time (after start) needed to complete the
    /// send.
    boost::shared_ptr<sstmac::transfer_time> send_time_;

    /// Timing model for time (after start) of transit of data in the
    /// network.
    boost::shared_ptr<sstmac::transfer_time> tran_time_;

    /// Timing model for time (after start) needed to complete the
    /// recv.
    boost::shared_ptr<sstmac::transfer_time> recv_time_;

    /// The network topology.  This object should eventually take over the
    /// ownership of per-link latencies and bandwidths.
    boost::shared_ptr<sstmac::network> network_;

    /// Reservation queues for each of the edges in our network.
    /// It is the sstmac::network's job to determine whether edges
    /// are shared in both directions (half duplex) or not (full duplex).
    std::vector<edge> edge_;

    /// The event manager.
    boost::shared_ptr<eventmanager> mgr_;

    /// The latency passed to the ctor. This and the bandwidth
    /// are used to construct timing models.
    timestamp latency_;

    /// The bandwidth passed to the ctor. This and the latency
    /// are used to construct timing models.
    bandwidth bandwidth_;

  protected:
    /// Hi.
    circuit2(timestamp latency, const bandwidth &bwidth,
             const boost::shared_ptr<sstmac::network> &net);

    /// Who am I?
    void set_self(const boost::shared_ptr<circuit2> &self);

    /// This is who I am.
    boost::shared_ptr<circuit2> get_self();

  public:
    /// New circuit2 network.
    static boost::shared_ptr<circuit2>
    construct(const timestamp latency, const bandwidth &bwidth,
              const boost::shared_ptr<sstmac::network> &net);

    /// Goodbye.
    virtual ~circuit2() throw();

    /// Required method from parent.
    virtual int slots() const;

    /// My event manager.
    virtual void set_eventmanager(const boost::shared_ptr<eventmanager> &mgr);

    /// Send the given message at the given time.
    /// Notify the sender node when the send is complete
    /// and notify the receiver node when receive needs to be performed.
    virtual void send(const timestamp &start_time,
                      const data &payload,
                      nodeid sender,
                      nodeid recver,
                      const boost::shared_ptr<eventhandler> &sendhandler,
                      const boost::shared_ptr<eventhandler> &recvhandler);

    /// Use the original timing model.
    void original_timing_model();

    /// Use the split timing model.
    void split_timing_model();
  };

} // end of namespace sstmac

#endif // ! SSTMAC_SIMULATOR_CIRCUIT2_H
