/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SIMULATOR_DATA_H
#define SSTMAC_SIMULATOR_DATA_H

#include <sstmac/simulator/timestamp.h>
#include <map>
#include <string>
#include <stdint.h>

namespace sstmac {

  /**
   * A container to store a fixed-point representation of data.
   *
   * By, the value is stored as a 64-bit integer
   * rather than an arbitrary precision value.  To support versions that
   * use arbitrary precision integers, you should not assume that
   * data_t has a native integer representation.
   */
  class data {
  public:
#ifndef SSTMAC_USE_GMPXX
    /// The type used to represent data in bits.
    typedef uint64_t data_t;
#else
    typedef timestamp::tick_t data_t;
#endif

    /// The units recognized for data sizes.
    ///
    /// K or Ki (for 10**3  and 2**10, respectively),
    /// M or Mi (for 10**6  and 2**20, respectively),
    /// G or Gi (for 10**9  and 2**30, respectively),
    /// T or Ti (for 10**12 and 2**40, respectively),
    /// P or Pi (for 10**15 and 2**50, respectively).
    /// E or Ei (for 10**18 and 2**60, respectively), or
    ///
    /// Unit type is either b (for bits) or B (for bytes).
    ///
    /// Unfortunately, swig cannot handle unsigned 64-bit enums,
    /// so we cannot stash the conversion factors in here.
    enum units {
      b=0, bits=b,  Kb, Kib, Mb, Mib, Gb, Gib, Tb, Tib, Pb, Pib, Eb, Eib,
      B,   Bytes=B, KB, KiB, MB, MiB, GB, GiB, TB, TiB, PB, PiB, EB, EiB
    };

  private:
    /// Map string names to bandwidth units.
    typedef std::map<std::string, units> unitmap_t;
    static unitmap_t *unitmap_;

    /// Initialize the unit map.
    static void init_unitmap();

    /// The input data (stored in bits).
    data_t data_;

    /// Private method to range check an input value.
    void set_size(double quantity, units units);

  public:
    /// Default constructor -- build a zero-size data object.
    data();

    /// Build a data object with exactly the given number of bits.
    data(const data_t &size);

    /// Build a new data object.
    data(double quantity, units unit);

    /// Alternate convenience way of defining the units (as a string).
    data(double quantity, const std::string &unit);

    /// Access the exact number of bits we're working with.
    const data_t& get_bits() const {
      return data_;
    }

    /// Return the data size in the desired units.
    double get_data(units unit) const;

    /// A convenience function to specify units using a string value.
    double get_data(const std::string &unit) const;
  };

  data operator+(const data &a, const data &b);
  data operator-(const data &a, const data &b);
  data operator*(const data &t, double scaling);
  data operator*(double scaling, const data &t);
  data operator/(const data &t, double scaling);

  std::ostream& operator<<(std::ostream &os, const data &t);

  inline bool operator<(const data &a, const data &b) {
    return (a.get_bits() < b.get_bits());
  }

  inline bool operator<=(const data &a, const data &b) {
    return (a.get_bits() <= b.get_bits());
  }

  inline bool operator>(const data &a, const data &b) {
    return (a.get_bits() > b.get_bits());
  }

  inline bool operator>=(const data &a, const data &b) {
    return (a.get_bits() >= b.get_bits());
  }

  inline bool operator==(const data &a, const data &b) {
    return (a.get_bits() == b.get_bits());
  }

  inline bool operator!=(const data &a, const data &b) {
    return (a.get_bits() != b.get_bits());
  }

} // end of namespace sstmac.

#endif // ! SSTMAC_SIMULATOR_DATA_H
