/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SIMULATOR_FULLNETWORK_H
#define SSTMAC_SIMULATOR_FULLNETWORK_H

#include <sstmac/simulator/interconnect.h>
#include <sstmac/simulator/bandwidth.h>
#include <boost/weak_ptr.hpp>
#include <map>

namespace sstmac {

  // Forward declaration of important types.
  class latencymodel;

  /** The fullnetwork uses an idealized network model that only
      considers the latency and bandwidth. The network is
      completely congestion free. The input for fullnetwork
      can specify the following attributes:
      
      <table border="1">
      
      <tr><td>Keyword<td>Type<td>Default<td>Description
      
      <tr><td><tt>bandwidth</tt><td>double<td>1000000000<td>The network
      bandwidth in bytes per second.
      
      <tr><td><tt>latency</tt><td>double<td>0.000001<td>The network
      latency in seconds.
      
      <tr><td><tt>nnode</tt><td>long<td>0<td>The number of nodes in
      the network.  The default of zero means that any number of nodes
        is supported.
        
        </table>
  */
  class fullnetwork: public interconnect {
  private:
    /// The latency of the network.
    boost::shared_ptr<latencymodel> latency_;
    /// The bandwidth of the network in bytes per second (why not bits?).
    bandwidth bandwidth_;
    /// The maximum size of the network.
    long size_;

    /// My event manager.
    boost::shared_ptr<eventmanager> eventmanager_;

    /// This is who I am.
    boost::weak_ptr<fullnetwork> self_;

  protected:
    /// Construct a fullnetwork object with the given properties
    /// \param latency is the network latency model
    /// \param bandwidth is the network bandwidth (bytes/second)
    /// \param maxsize is the maximum network size (<= 0 for unlimited size).
    fullnetwork(const boost::shared_ptr<latencymodel> &latency,
                bandwidth bw, int maxsize);

    /// Tell me who I am.
    void set_self(const boost::shared_ptr<fullnetwork>&);

    /// Who am I?
    boost::shared_ptr<fullnetwork> get_self();

  public:
    /// Construct a full network with fixed latency (legacy interface).
    static boost::shared_ptr<fullnetwork>
    construct(const timestamp &latency, double bandwidth, int maxsize = -1);

    /// Construct a full network with caller-defined latency.
    static boost::shared_ptr<fullnetwork>
    construct(const boost::shared_ptr<latencymodel> &latency,
              bandwidth bw, int maxsize = -1);

    /// Exit stage left.
    virtual ~fullnetwork() throw();

    /// The maximum number of nodes possible in this simulator.
    /// A value of <= 0 indicates that any number of nodes is possible.
    virtual int slots() const;

    /// Define the event manager.
    virtual void set_eventmanager(const boost::shared_ptr<eventmanager> &mgr);

    /// Send the given message at the given time.
    /// Notify the sender node when the send is complete
    /// and notify the receiver node when receive needs to be performed.
    virtual void send(const timestamp &start_time,
                      int64_t bytes,
                      nodeid sender,
                      nodeid recver,
                      const boost::shared_ptr<eventhandler> &sendhandler,
                      const boost::shared_ptr<eventhandler> &recvhandler);
  };

} // end of namespace sstmac

#endif // ! SSTMAC_SIMULATOR_FULLNETWORK_H
