/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SIMULATOR_KERNEL_H
#define SSTMAC_SIMULATOR_KERNEL_H

#include <sstmac/simulator/timestamp.h>
#include <sstmac/util/nocopy.h>
#include <boost/shared_ptr.hpp>
#include <boost/weak_ptr.hpp>
#include <list>

namespace sstmac {

  // Forward declaration of important types.
  class eventhandler;
  class nodeinterface;

  /**
   * A handle on a kernel.
   */
  class kernel : virtual public nocopy {
    /// Tag whether this kernel has been completed.
    bool done_;
    /// Any eventhandlers waiting for my completion.
    typedef std::list< boost::shared_ptr<eventhandler> > complist_t;
    complist_t complist_;

  protected:
    /// Hello.
    kernel();

  public:
    /// Virtual destructor is essential.
    virtual ~kernel() throw();

    /// Respond to an event call.
    /// \param start_time is the current simulation time.
    /// \param my_node is the nodedelegate that bridges this kernel to a node.
    /// \throw illformederror if you don't know how to handle this event.
    virtual void start(const timestamp& start_time,
                       const boost::shared_ptr<nodeinterface> &my_node) = 0;

    /// Test whether this kernel has been completed.
    bool is_complete() const;

    /// Notify an eventhandler upon the completion of this kernel.
    /// \throw illformederror if the kernel is complete (is_complete() is true).
    void add_listener(const boost::shared_ptr<eventhandler>&);

    /// Un-register a callback for completion notification.
    void erase_listener(const boost::shared_ptr<eventhandler>&);

    /// Mark this kernel as complete.  Callback used by kernelcompletion.
    /// This method notifies all listeners, so don't forget to call
    /// kernel::complete if you override this method.
    virtual void complete(const timestamp &current_time);
  };

} // end of namespace sstmac

#endif // ! SSTMAC_SIMULATOR_KERNEL_H
