/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/simulator/nodeinterface.h>
#include <sstmac/simulator/node.h>
#include <sstmac/process/server.h>
#include <sstmac/simulator/kernel.h>
#include <sstmac/simulator/kernelhandler.h>
#include <sstmac/processor/instructions.h>
#include <sstmac/util/errors.h>

namespace sstmac {

  // Pointer types.
  using boost::shared_ptr;
  using boost::weak_ptr;

  //
  // Hi there.
  //
  nodeinterface::nodeinterface(const boost::shared_ptr<node> &my_node,
                               const ppid &my_ppid,
                               const taskid &my_taskid) :
    node_(my_node), mgr_(my_node->get_eventmanager()),
    ppid_(my_ppid), taskid_(my_taskid)
  {}

  //
  // Who am I?
  //
  void nodeinterface::set_self(const shared_ptr<nodeinterface> &self) {
    if(self.get() != this) {
      throw valueerror("nodeinterface::set_self:  'self' pointer does not "
                       "point to 'this' object.");
    }
    self_ = self;
  }

  //
  // This is who I am.
  //
  shared_ptr<nodeinterface> nodeinterface::get_self() {
    return self_.lock();
  }

  //
  // Hello.
  //
  shared_ptr<nodeinterface>
  nodeinterface::construct(const boost::shared_ptr<node> &mynode,
                           const ppid &myppid,
                           const taskid& mytid)
  {
    shared_ptr<nodeinterface> retval(new nodeinterface(mynode, myppid, mytid));
    retval->set_self(retval);
    return retval;
  }

  //
  // Goodbye.
  //
  nodeinterface::~nodeinterface() throw() 
  {}

  //
  // Get the ppid associated with this parallel simulation.
  //
  const ppid& nodeinterface::get_ppid() const {
    return ppid_;
  }

  //
  // Get the taskid associated with this peer instance.
  //
  const taskid& nodeinterface::get_taskid() const {
    return taskid_;
  }

  //
  // Get current time from the node.
  //
//   const timestamp& nodeinterface::now() const {
//     return node_.lock()->now();
//   }

  //
  // Test whether a server is defined.
  //
  bool nodeinterface::has_server() {
    return node_.lock()->has_server(ppid_);
  }

  //
  // Set a network event server for the given ppid.
  //
  void nodeinterface::set_server(const shared_ptr<server> &the_server) {
    node_.lock()->set_server(ppid_, the_server);
  }

  //
  // Access the server.
  //
  shared_ptr<server> nodeinterface::get_server() {
    return node_.lock()->get_server(ppid_);
  }

  //
  // Set a simple trigger (wake-up call) for this kernel.
  //
  void nodeinterface::schedule(const timestamp &start_time,
                               const boost::shared_ptr<eventhandler> &event)
  {
    node_.lock()->schedule(start_time, ppid_, event);
  }

  //
  // Start a new kernel at the given time.
  //
  void nodeinterface::add(const timestamp &start_time,
                          const boost::shared_ptr<kernel> &task)
  {
    // Delegate the startup to a new kernelhandler.
    shared_ptr<eventhandler> ev(kernelhandler::construct(task, self_.lock()));
    node_.lock()->schedule(start_time, ppid_, ev);
  }

  //
  // Pretend to compute.
  //
  void nodeinterface::compute(const timestamp &start_time,
                              const timestamp &cputime,
                              const boost::shared_ptr<eventhandler> &completion)
  {
    node_.lock()->compute(start_time, ppid_, cputime, completion);
  }

  //
  // Pretend to compute on multiple threads.
  //
  void nodeinterface::compute(const timestamp &start_time,
                              const timestamp &cputime_per_thread,
                              int threads,
                              const boost::shared_ptr<eventhandler> &completion)
  {
    node_.lock()->compute(start_time, ppid_, cputime_per_thread,
                          threads, completion);
  }

  //
  // I am not sure yet how best to handle this one :)
  //
  void nodeinterface::compute(const timestamp &start_time,
                              const boost::shared_ptr<instructions> &instr,
                              const boost::shared_ptr<eventhandler> &completion)
  {
    node_.lock()->compute(start_time, ppid_, instr, completion);
  }

  //
  // Request a copy of the given number of bytes between two address spaces.
  //
  void nodeinterface::copy(const timestamp &start_time,
                           int64_t bytes,
                           const boost::shared_ptr<eventhandler> &completion)
  {
    node_.lock()->copy(start_time, ppid_, bytes, completion);
  }

} // end of namespace sstmac
