/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SIMULATOR_TYPICAL_FLOWER_H
#define SSTMAC_SIMULATOR_TYPICAL_FLOWER_H

#include <sstmac/simulator/flower.h>
#include <sstmac/simulator/netsharing.h>
#include <sstmac/simulator/timestamp.h>
#include <sstmac/network/network.h>
#include <vector>
#include <map>
#include <set>
#include <string>
#include <boost/shared_ptr.hpp>

namespace sstmac {
  
  /**
   * A base class for flowers that explicity track flows and edge
   * bandwidth utilization.
   */
  class typical_flower: public flower {
  protected:
    typedef std::map<flow_id_t, double> flow_bw_t;

  private:
    flow_bw_t flow_bw_;
    std::vector<double> edge_bw_utilization_;

  protected:
    flow_id_t next_flow_id_;
    flow_id_t stopping_flow_;
    timestamp stopping_time_;

    double max_bw_;
    std::vector<double> edge_bw_;
    double edge_beta(size_t i) const { return 1.0/edge_bw_.at(i); }
    double edge_bw(size_t i) const { return edge_bw_.at(i); }
    /// current time for flow data
    timestamp t_;
    //
    typedef std::map<flow_id_t, std::vector<edge_id_t> > flow_edges_t;
    flow_edges_t flow_edges_;
    //
    typedef std::set<flow_id_t> flow_set_t;
    typedef std::vector<flow_set_t> edge_flows_t;
    edge_flows_t edge_flows_;
    /// bytes remaining in flow
    typedef std::map<flow_id_t,double> flow_nbyte_t;
    flow_nbyte_t flow_nbyte_; 

    /// The maximum allowable error in bandwidth error checks.
    double bw_epsilon_;

    void advance_time(timestamp);
#undef TYPICAL_FLOWER_USE_OLD_EDGE_BW_UTIL
#ifndef TYPICAL_FLOWER_USE_OLD_EDGE_BW_UTIL
    double edge_bw_utilization(edge_id_t e) { return edge_bw_utilization_[e]; }
#else // TYPICAL_FLOWER_USE_OLD_EDGE_BW_UTIL
    double edge_bw_utilization(edge_id_t e);
#endif // TYPICAL_FLOWER_USE_OLD_EDGE_BW_UTIL

    double edge_bw_equal_and_higher_priority_utilization(edge_id_t e, flow_id_t f);
    void find_next_stop();
    void set_flow_bw(flow_id_t fid, double bw);
    void add_to_flow_bw(flow_id_t fid, double delta_bw);
    void scale_flow_bw(flow_id_t fid, double scaling_factor);
    double get_flow_bw(flow_id_t);
    void erase_flow_bw(flow_id_t);
    const flow_bw_t &flow_bw() const { return flow_bw_; }

    typical_flower(const boost::shared_ptr<sstmac::network> &net,
           double node_router_bw, double router_router_bw);

    /// Start the given flow, rebalancing other flows as required.
    /// This is only called for new flows.
    virtual void rebalance_flow_start(flow_id_t f) = 0;

    /// Stop the given flow, rebalancing other flows as required.
    virtual void rebalance_flow_stop(flow_id_t oldf) = 0;

  public:
    virtual ~typical_flower() throw() {}

    /// Override flower::start_flow.
    flow_id_t start_flow(timestamp tstart, const std::vector<edge_id_t> &edges, long nbyte);

    /// Override flower::stop_flow.
    void stop_flow(timestamp t, flow_id_t f);

    /// Override flower::next_stopping_time.
    timestamp next_stopping_time();

    /// Override flower::next_stopping_flow.
    flow_id_t next_stopping_flow();
  };

} // end of namespace sstmac

#endif // ! SSTMAC_SIMULATOR_TYPICAL_FLOWER_H
