/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009-2010 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SKELETON_ACTORMATMUL_ACTORMATMUL_COMPUTOR_H
#define SSTMAC_SKELETON_ACTORMATMUL_ACTORMATMUL_COMPUTOR_H

#include <sstmac/skeleton/actormatmul/actormatmul.h>
#include <sstmac/skeleton/actormatmul/actormatmul-messages.h>

namespace sstmac {

  /**
   * An actor that performs computations (even actors have to make a living).
   */
  class actormatmul::computor : public simplemboxapp {
    typedef actormatmul::localstore localstore;
    typedef actormatmul::remotestore remotestore;
    typedef actormatmul::work work;
    typedef actormatmul::putdata putdata;

    /// The parent id.
    actorid parent_;
    /// The number of threads managed by this computor.
    int threads_;
    /// The performance.
    timestamp time_per_fma_;

    /// Private callback to pretend to do computation.
    void compute(const boost::shared_ptr<work>& msg) {
      int64_t md = msg->matdim();
      int64_t multiplies = md*md*md;
      mbox()->compute((time_per_fma_ * multiplies)/threads_, threads_);
    }

    /// Hello.
    computor(actorid parent, int threads, const timestamp &time_per_FMA) :
      parent_(parent), threads_(threads), time_per_fma_(time_per_FMA)
    {}

  public:
    /// Create a new computor.
    static boost::shared_ptr<computor>
    construct(actorid parent, int threads, const timestamp &per_FMA) {
      boost::shared_ptr<computor> retval
        (new computor(parent, threads, per_FMA));
      retval->set_self(retval);
      return retval;
    }

    /// Clone this object.
    boost::shared_ptr<simplemboxapp> clone() const {
      return construct(parent_, threads_, time_per_fma_);
    }

    /// Go.
    void run() {
      do {
        simplembox::recvresult_t res =
          mbox()->recv(actorid::any(), actorpattern::any());
        boost::shared_ptr<work> msg =
          boost::dynamic_pointer_cast<work>(res.second);
        if(msg->matdim() > 0) {
          this->compute(msg);
          if(msg->store_to() == parent_) {
            // We send back 8 bytes -- we are assuming we can share pointers
            mbox()->send(parent_, localstore::construct(msg->iteration()), 8);
          }
          else {
            // remote store.  We have to send back a copy of the actual matrix.
            mbox()->send(msg->store_to(),
                         remotestore::construct(msg->iteration()),
                         msg->matdim()*msg->matdim());
          }
        }
        else
          break;
      } while(true);
    }
  };


} // end of namespace sstmac

#endif // ! SSTMAC_SKELETON_ACTORMATMUL_ACTORMATMUL_COMPUTOR_H
