/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/skeleton/mpi3d/mpi3d.h>
#include <sstmac/process/mpistatus.h>
#include <sstmac/process/mpitype.h>
#include <sstmac/process/mpitag.h>
#include <sstmac/simulator/nodeinterface.h>
#include <sstmac/util/debug.h>
#include <stdio.h>
#include <string.h>

namespace sstmac {

  std::ostream &
  operator << (std::ostream&o,mpi3d::coor&c)
  {
    o<<'{'<<c[0]<<','<<c[1]<<','<<c[2]<<'}';
    return o;
  }

  // Pointer types.
  using boost::shared_ptr;

  //
  // Hi.
  //
  mpi3d::mpi3d(long nxa, long nya, long nza) :
    compute_time(0), noiselevel(0.1), rngseed(0xF00D)
  {
    l = 100;
    g = 1;
    nfield = 1;
    nbyte = 8;
    nx = nxa;
    ny = nya;
    nz = nza;
    gamma = 100e-9;
    verbose = false;
    nstep = 10;
    waitall = true;
  }

  //
  // Hello.
  //
  shared_ptr<mpi3d> mpi3d::construct(long nx, long ny, long nz) {
    //std::cerr << "sizeof(mpi3d)==" << sizeof(mpi3d) << "\n";
    shared_ptr<mpi3d> obj(new mpi3d(nx,ny,nz));
    obj->set_self(obj);
    return obj;
  }

  //
  // Goodbye.
  //
  mpi3d::~mpi3d() throw() {
  }

  //
  // Get a copy.
  //
  shared_ptr<mpiapp> mpi3d::clone() {
    shared_ptr<mpi3d> app = mpi3d::construct(nx,ny,nz);
    app->verbose = verbose;
    app->l = l;
    app->g = g;
    app->nfield = nfield;
    app->nbyte = nbyte;
    app->gamma = gamma;
    app->nx = nx;
    app->ny = ny;
    app->nz = nz;
    app->nstep = nstep;
    app->waitall = waitall;
    app->compute_time = compute_time;
    app->noiselevel = noiselevel;
    app->rngseed = rngseed;
    return app;
  }

  //
  // Set compute time targets.
  //
  void mpi3d::set_compute(const timestamp &ts, double noise,
                          unsigned long seed)
  {
    compute_time = ts;
    noiselevel = noise;
    rngseed = seed;
  }

  //
  // Go.
  //
  void mpi3d::run() {
    timestamp start = mpi()->init();
    mpicomm world = mpi()->comm_world();
    comm_ = world;
    mpiid rank = world.rank();
    mpiid size = world.size();
    mpitag tag(0);

    if (rank.id >= nx*ny*nz) {
      mpi()->finalize();
      return;
    }

    if(compute_time != timestamp(0)) {
      rng.seed(rngseed + rank.id);
    }

    init();

    std::vector<mpirequest_t> irecv_reqs;
    std::vector<mpirequest_t> isend_reqs;

    if(rank.id == 0) {
      if(debugger::active()) debug() << "mpi3d::run():  " << face_neighbors_.size() << " face neighbors, " << edge_neighbors_.size() << " edge neighbors, " << corner_neighbors_.size() << " corner neighbors\n";
    }
    for (long i=0; i<nstep; i++) {
      if(debugger::active()) debug() << "mpi3d::run: {" << rank << "} :  Iteration " << i << " of " << nstep << "\n";
      irecv_reqs.clear();
      isend_reqs.clear();

      if(debugger::active()) debug() << "mpi3d::run: {" << rank << "} :  running irecv commands\n";
      run_irecv(face_nbyte_, face_neighbors_, irecv_reqs);
      //run_irecv(edge_nbyte_, edge_neighbors_, irecv_reqs);
      //run_irecv(corner_nbyte_, corner_neighbors_, irecv_reqs);

      if(debugger::active()) debug() << "mpi3d::run: {" << rank << "} :  running isend commands\n";
      run_isend(face_nbyte_, face_neighbors_, isend_reqs);
      //run_isend(edge_nbyte_, edge_neighbors_, isend_reqs);
      //run_isend(corner_nbyte_, corner_neighbors_, isend_reqs);
      if (verbose) {
        std::cout << "waitall 1" << std::endl;
      }
      if(debugger::active()) debug() << "mpi3d::run: {" << rank << "} :  waiting for irecv commands\n";
      run_waitall(irecv_reqs);
      //run_compute();
      if (verbose) {
        std::cout << "waitall 2" << std::endl;
      }
      if(debugger::active()) debug() << "mpi3d::run: {" << rank << "} :  waiting for isend commands\n";
      run_waitall(isend_reqs);
      if(debugger::active()) debug() << "mpi3d::run: {" << rank << "} :  Bottom of iterative loop\n";
    }

    if(debugger::active()) debug() << "mpi3d::run: {" << rank << "} :  finalizing mpi\n";
    timestamp done = mpi()->finalize();

    double runtime = (done-start).sec();
    /*
    if(rank.id == 0) {
      static const size_t bsize = 1023;
      char *buffer = new char[bsize+1];
      snprintf(buffer, bsize, "mpi3d run started at %f and finished at %f "
               "with a total runtime of %f seconds\n",
               start.sec(), done.sec(), runtime);
      buffer[bsize] = '\0'; // just to be sure.
      if(! fwrite(buffer, sizeof(char), strlen(buffer), stderr)) {
        // Failed to write the buffer -- but there isn't much we can do about it
      }
      delete[] buffer;
//     if(debugger::active()) debug() << "Peer " << rank << " started at " << start
//               << " and finished at " << done << ", with a total runtime of "
//               << runtime << " seconds\n";
    }
    */
  }

  mpi3d::coor
  mpi3d::rank_to_coor(mpiid rank)
  {
    coor ret;
    ret[0] = rank.id/(ny*nz);
    ret[1] = (rank.id-ret[0]*ny*nz)/nz;
    ret[2] = rank.id-ret[0]*ny*nz-ret[1]*nz;
    return ret;
  }

  mpiid
  mpi3d::coor_to_rank(const coor &c)
  {
    return mpiid(c[0]*ny*nz + c[1]*nz + c[2]);
  }

  mpiid
  mpi3d::offset_to_rank(long dx, long dy, long dz)
  {
    return coor_to_rank(coor(my_coor_[0]+dx,my_coor_[1]+dy,my_coor_[2]+dz));
  }

  void
  mpi3d::run_irecv(long nbyte, const std::vector<mpiid> &neighbors,
                   std::vector< mpirequest_t > &reqs)
  {
    for (unsigned long i=0; i < neighbors.size(); i++) {
      mpirequest_t req;
      mpi()->irecv(nbyte, mpitype::mpi_double, neighbors[i],
                tag_, comm_, req);
      reqs.push_back(req);
      // Make the irecv call "not completely free"
      // This may later be replaced with a more reasonable approach
      //mpi()->compute(timestamp::exact_nsec(1));
    }
  }

  void
  mpi3d::run_isend(long nbyte, const std::vector<mpiid> &neighbors,
                   std::vector< mpirequest_t > &reqs)
  {
    for (unsigned long i=0; i < neighbors.size(); i++) {
      mpirequest_t req;
      mpi()->isend(nbyte, mpitype::mpi_double, neighbors[i],
                tag_, comm_, req);
      reqs.push_back(req);
      // Make the isend call "not completely free"
      // This may later be replaced with a more reasonable approach
      //mpi()->compute(timestamp::exact_nsec(1));
    }
  }

  void
  mpi3d::run_waitall(std::vector< mpirequest_t > &reqs)
  {
    if (waitall) {
      std::vector<const_mpistatus_t> stats(reqs.size());
      if (reqs.size() > 0) {
        mpi()->waitall(reqs, stats);
      }
    }
    else {
      const_mpistatus_t stat;
      for (size_t i=0; i<reqs.size(); i++) {
        if(debugger::active()) debug() << "mpi3d[" << this << "]::run_waitall:  Waiting for request "
                  << reqs[i] << " (" << i << " of " << reqs.size()-1 << "\n";
        mpi()->wait(reqs[i],stat);
        if(debugger::active()) debug() << "mpi3d[" << this << "]::run_waitall:  Request "
                  << i << " of " << reqs.size()-1 << " complete\n";
      }
    }
  }

  void
  mpi3d::run_compute()
  {
    if(compute_time != timestamp(0)) {
      if (noiselevel != 0) {
        timestamp delay = compute_time * noiselevel * rng();
        mpi()->compute(compute_time + delay);
      }
      else {
        mpi()->compute(compute_time);
      }
    }
  }

  void
  mpi3d::init_faces()
  {
    std::vector<long> ngrid(3);
    ngrid[0] = nx;
    ngrid[1] = ny;
    ngrid[2] = nz;
    face_nbyte_ = g*l*l*nfield*nbyte;
    face_neighbors_.clear();
    face_neighbors_.reserve(6);
    for (long i=-1; i<=1; i+=2) {
      for (long j=0; j<3; j++) {
        coor neighbor(my_coor_);
        neighbor[j] = (neighbor[j] + ngrid[j] + i)%ngrid[j];
        mpiid neighbor_rank = coor_to_rank(neighbor);
        if (neighbor_rank != my_rank_) {
          if (verbose) {
            std::cout << "face: "
                      << my_rank_ << " -> " << neighbor_rank
                      << " " << my_coor_ << " -> " << neighbor
                      << std::endl;
          }
          face_neighbors_.push_back(neighbor_rank);
        }
      }
    }
  }

  void
  mpi3d::init_edges()
  {
    std::vector<long> ngrid(3);
    ngrid[0] = nx;
    ngrid[1] = ny;
    ngrid[2] = nz;
    edge_nbyte_ = g*g*l*nfield*nbyte;
    edge_neighbors_.clear();
    edge_neighbors_.reserve(12);
    for (long i=-1; i<=1; i+=2) {
      for (long j=-1; j<=1; j+=2) {
        long cases[3][2] = { { 0, 1}, {0, 2}, {1, 2} };
        for (long k=0; k<3; k++) {
          long a = cases[k][0];
          long b = cases[k][1];
          coor neighbor(my_coor_);
          neighbor[a] = (neighbor[a] + ngrid[a] + i)%ngrid[a];
          neighbor[b] = (neighbor[b] + ngrid[b] + j)%ngrid[b];
          mpiid neighbor_rank = coor_to_rank(neighbor);
          if (neighbor_rank != my_rank_) {
            if (verbose) {
              std::cout << "edge: "
                        << my_rank_ << " -> " << neighbor_rank
                        << " " << my_coor_ << " -> " << neighbor
                        << std::endl;
            }
            edge_neighbors_.push_back(neighbor_rank);
          }
        }
      }
    }
  }

  void
  mpi3d::init_corners()
  {
    std::vector<long> ngrid(3);
    ngrid[0] = nx;
    ngrid[1] = ny;
    ngrid[2] = nz;
    corner_nbyte_ = g*g*g*nfield*nbyte;
    corner_neighbors_.clear();
    corner_neighbors_.reserve(8);
    for (long i=-1; i<=1; i+=2) {
      for (long j=-1; j<=1; j+=2) {
        for (long k=-1; k<=1; k+=2) {
          coor neighbor(my_coor_);
          neighbor[0] = (neighbor[0] + ngrid[0] + i)%ngrid[0];
          neighbor[1] = (neighbor[1] + ngrid[1] + j)%ngrid[1];
          neighbor[2] = (neighbor[2] + ngrid[2] + k)%ngrid[2];
          mpiid neighbor_rank = coor_to_rank(neighbor);
          if (neighbor_rank != my_rank_) {
            if (verbose) {
              std::cout << "corner: "
                        << my_rank_ << " -> " << neighbor_rank
                        << " " << my_coor_ << " -> " << neighbor
                        << std::endl;
            }
            corner_neighbors_.push_back(neighbor_rank);
          }
        }
      }
    }
  }

  void
  mpi3d::init()
  {
    my_rank_ = comm_.rank();
    mpiid nproc = comm_.size();

    long nxyz = nx*ny*nz;
    if (nproc.id < nxyz) {
      std::cout << "mpi3d: given " << nproc
                << " processes but need " << nxyz << std::endl;
      abort();
    }

    tag_.id = 0;

    if (my_rank_.id < nxyz) {
      my_coor_ = rank_to_coor(my_rank_);
      // set up info need for face transfers
      init_faces();
      // set up info needed for edge transfers
      init_edges();
      // set up info needed for edge transfers
      init_corners();
    }
  }

} // end of namespace sstmac.
