/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SKELETON_MPI3D_MPI3D_H
#define SSTMAC_SKELETON_MPI3D_MPI3D_H

#include <sstmac/process/mpiapp.h>
#include <sstmac/process/mpiapi.h>
#include <vector>
#include <map>
#include <sstmac/process/mpiapp.h>
#include <math.h>
#include <stdlib.h>

namespace sstmac {

  /**
   * Basic MPI ping-pong.
   */
  class mpi3d : public mpiapp {
  public:
    typedef mpiapi::mpirequest_t mpirequest_t;
    typedef mpiapi::mpistatus_t mpistatus_t;
    typedef mpiapi::const_mpistatus_t const_mpistatus_t;

    // User adjustable fields.
    bool verbose;
    long l;
    long g;
    long nfield;
    long nbyte;
    double gamma;
    long nx;
    long ny;
    long nz;
    long nstep;
    bool waitall;

    // Helgi hacks.
    timestamp compute_time;
    double noiselevel;
    unsigned long rngseed;

  private:
    class gaussnoise {
      int valid;
      double val[2];
      unsigned int seed_;
      // Can you believe that boost::mt19937 is 5011 bytes in size!
    public:
      gaussnoise() : valid(0) {}
      void seed(unsigned int s) { seed_ = s; }
      double operator()() {
        if(!valid) {
          get_values();
        }
        return val[--valid];
      }
    private:
      void get_values() {
        // from http://www.taygeta.com/random/gaussian.html
        // Algorithm by Dr. Everett (Skip) Carter, Jr.
        double x1, x2, w;
        double max = RAND_MAX;
        do {
          x1 = 2.0 * (rand_r(&seed_) / max) - 1.0;
          x2 = 2.0 * (rand_r(&seed_) / max) - 1.0;
          w = x1 * x1 + x2 * x2;
        } while ( w >= 1.0 );
        w = sqrt( (-2.0 * log( w ) ) / w );
        val[0] = x1 * w;
        val[1] = x2 * w;
        valid = 2;
      }
    };
    gaussnoise rng;

  public: // must be public for operator <<
#ifndef SWIG // Do not parse nested classes with swig.
    /// This class represents a coordinate in the 3D grid.
    class coor: public std::vector<long> {
    public:
    coor(): std::vector<long>(3) {}
    coor(long x, long y, long z): std::vector<long>(3) {
        operator[](0) = x;
        operator[](1) = y;
        operator[](2) = z;
      }
    };
#endif // SWIG

  private:
    // These fields keep track of fixed information about the run.
    std::map<coor, long> coor_to_rank_;
    mpiid my_rank_;
    coor my_coor_;
    std::vector<mpiid> face_neighbors_;
    std::vector<mpiid> edge_neighbors_;
    std::vector<mpiid> corner_neighbors_;
    long face_nbyte_;
    long edge_nbyte_;
    long corner_nbyte_;
    mpitag tag_;
    mpicomm comm_;
    double time_;

    // utility functions
    coor rank_to_coor(mpiid rank);
    mpiid coor_to_rank(const coor &c);
    mpiid offset_to_rank(long dx, long dy, long dz);
    void run_irecv(long nbyte, const std::vector<mpiid> &neighbors,
                   std::vector<mpirequest_t> &reqs);
    void run_isend(long nbyte, const std::vector<mpiid> &neighbors,
                   std::vector<mpirequest_t> &reqs);
    void run_waitall(std::vector<mpirequest_t> &reqs);
    void run_compute();
    void init_faces();
    void init_edges();
    void init_corners();
    void init();

  protected:
    /// Hi.
    mpi3d(long nx, long ny, long nz);

  public:
    /// Hello.
    static boost::shared_ptr<mpi3d> construct(long nx, long ny, long nz);

    /// Goodbye.
    virtual ~mpi3d() throw();

    /// Get a copy.
    boost::shared_ptr<mpiapp> clone();

    void set_compute(const timestamp &ts, double noise,
                     unsigned long seed = 0xBADF00D);

    /// Go.
    virtual void run();
  };

#ifndef SWIG // mpi3d::coor is ifdef'ed out for swig above
  std::ostream & operator << (std::ostream&o,mpi3d::coor&c);
#endif // SWIG

} // end of namespace sstmac

#endif // ! SSTMAC_SKELETON_MPI3D_MPI3D_H
