/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SKELETON_MPIPARSERS_RSQMETATRACE_H
#define SSTMAC_SKELETON_MPIPARSERS_RSQMETATRACE_H

#include <sstmac/simulator/nodeid.h>
#include <sstmac/process/mpiid.h>
#include <boost/shared_ptr.hpp>
#include <vector>

namespace sstmac {

  /// Position in the network mesh.
  struct rsqcrd {
    int x, y, z;
    rsqcrd() : x(-1), y(-1), z(-1) {}
    rsqcrd(int xin, int yin, int zin) : x(xin), y(yin), z(zin) {}
    bool operator==(const rsqcrd &b) const {
      return ((x==b.x) && (y==b.y) && (z==b.z));
    }
    bool operator<(const rsqcrd &b) const {
      if(x != b.x) return (x < b.x);
      if(y != b.y) return (y < b.y);
      return z < b.z;
    }
  };

  /**
   * Meta-information regarding a collection of trace files gathered
   * on RSQ (if you don't know what this means, you probably don't need
   * to worry about this file -- lucky you).
   */
  class rsqmetatrace {
  private:
    /// The common file root for the input files.
    std::string fileroot_;

    /// The nodes requested so far.
    typedef std::vector<rsqcrd> node_crd_t;
    node_crd_t node_crd_;

    /// The positions of each of the mpi peers.
    typedef std::vector<rsqcrd> peer_crd_t;
    peer_crd_t peer_crd_;

  protected:
    /// Create a new metatrace with enough space for the given number of peers.
    rsqmetatrace(const std::string &fileroot, size_t peers);

  public:
    /// Hello.
    static boost::shared_ptr<rsqmetatrace> construct(const std::string &fileroot, size_t peers);

    /// Farewell.
    virtual ~rsqmetatrace() throw();

    /// Get the common file root for the input files.
    const std::string& fileroot() const;

    /// The number of nodes in the system.  Note that this will often
    /// be different than the number of peers.
    nodeid nodes() const;

    /// Get the coordinate of the nth node in the run.
    const rsqcrd& node_crd(const nodeid &node) const;

    /// The number of peers in this trace.
    mpiid peers() const;

    /// Get the coordinate of the nth peer in the run
    const rsqcrd& peer_crd(const mpiid &peer) const;

    /// Set the mesh coordinate of the nth node.
    /// Note that multiple peers can reside on the same node.
    void set_peer_crd(const mpiid &peer, const rsqcrd &crd);
  };

} // end of namespace sstmac

#endif // SSTMAC_SKELETON_MPIPARSERS_RSQMETATRACE_H
