/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/skeleton/osubenchmarks/osulat.h>
#include <sstmac/process/mpiapi.h>
#include <sstmac/process/mpistatus.h>
#include <sstmac/process/mpitype.h>
#include <sstmac/process/mpitag.h>
#include <sstmac/util/errors.h>
#include <sstmac/util/debug.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

namespace sstmac {

  //
  // Hi.
  //
  osulat::osulat(const std::string &ofile) :
    ofile_(ofile), max_req_num(1000), msg_alignment(64), max_msg_size(1<<22),
    loop(100),  skip(1000), loop_large(100),
    skip_large(10), large_message_size(8192)
  {}

  //
  // It's creation time.
  //
  boost::shared_ptr<osulat> osulat::construct(const std::string &ofile) {
    boost::shared_ptr<osulat> obj(new osulat(ofile));
    obj->set_self(obj);
    return obj;
  }

  //
  // Goodbye.
  //
  osulat::~osulat() throw() {
  }

  //
  // Get a copy.
  //
  boost::shared_ptr<mpiapp> osulat::clone() {
    boost::shared_ptr<osulat> myclone = construct(ofile_);
    myclone->max_req_num = this->max_req_num;
    myclone->msg_alignment = this->msg_alignment;
    myclone->max_msg_size = this->max_msg_size;
    myclone->loop = this->loop;
    myclone->skip = this->skip;
    myclone->loop_large = this->loop_large;
    myclone->skip_large = this->skip_large;
    myclone->large_message_size = this->large_message_size;
    return myclone;
  }

  //
  // Go.
  //
  void osulat::run() {
    int i, j, size;
    mpiid myid, numprocs;
    timestamp t_start, t_end, t;
    static const mpiid zeroid(0), oneid(1);
    int currloop = loop;

    FILE *fout = stderr;
    if(ofile_ != "") {
      fout = fopen(ofile_.c_str(), "w");
      if(! fout) {
        fprintf(stderr, "osbuw::run:  Failed to open %s for output: %s\n",
                ofile_.c_str(), strerror(errno));
        throw ioerror("osulat::run:  Failed to open output file\n");
      }
    }
    
    mpicomm world = mpi()->comm_world();
    mpi()->init();
    myid = world.rank();
    numprocs = world.size();

    //align_size = getpagesize();
    //assert(align_size <= MAX_ALIGNMENT);
    int align_size = msg_alignment;

    if(numprocs.id != 2) {
      if(myid == zeroid) {
        fprintf(stderr, "osulat: requires exactly two processes\n");
      }
      mpi()->finalize();
      return;
    }

    if(myid == zeroid) {
      fprintf(fout, "# %s %s\n", "SST/macro OSU latency", "v3.1.1");
      fprintf(fout, "%-9s %20s\n", "# Size", "Latency (us)");
      fflush(fout);
    }

    /* Latency test */
    for(size = 0; size <= max_msg_size; size = (size ? size *= 2 : size + 1)) {
      if(size > large_message_size) {
        currloop = loop_large;
        skip = skip_large;
      }
      mpi()->barrier(world);
      if(myid == zeroid) {
        for(i = 0; i < currloop + skip; i++) {
          if(i == skip) {
            t_start = mpi()->wtime();
          }
          std::vector<mpiapi::const_mpistatus_t> reqstat(1);
          mpi()->send(size,mpitype::mpi_char,oneid,mpitag(1),world);
          mpi()->recv(size,mpitype::mpi_char,oneid,mpitag(1),world,reqstat.at(0));
        }
        t_end = mpi()->wtime();
      } else if(myid == oneid) {
        for(i = 0; i < currloop + skip; i++) {
          std::vector<mpiapi::const_mpistatus_t> reqstat(1);
          mpi()->recv(size,mpitype::mpi_char,zeroid,mpitag(1),world,reqstat.at(0));
          mpi()->send(size,mpitype::mpi_char,zeroid,mpitag(1),world);
        }
      }

      if(myid == zeroid) {
        t = t_end - t_start;
        double latency = t.usec() / (2.0 * currloop);

        fprintf(fout, "%-9d %20.2f\n", size, latency);
        fflush(fout);
      }
    }
    
    mpi()->finalize();
  }

} // end of namespace sstmac.
