/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/util/stackalloc-chunk.h>
#include <sstmac/util/errors.h>
#include <sys/mman.h>
#include <unistd.h>
#include <iostream>
#include <stdio.h>
#include <string.h>
#include <errno.h>

namespace sstmac {

  //
  // Make a new chunk.
  //
  stackalloc::chunk::chunk(size_t stacksize, size_t suggested_chunk_size,
                           bool use_mprot) :
    addr_(0), size_(suggested_chunk_size),
    stacksize_(stacksize), use_mprot_(use_mprot), next_stack_(0)
  {
    // Figure out how big we want our chunk to be.
    const size_t page = sysconf(_SC_PAGESIZE);
    const size_t guard = (use_mprot_ ? page : 0);
    if(size_ < (stacksize_ + 2*guard)) {
      size_ = stacksize_ + 2*guard;
    }
    // Round size_ up to match a multiple of allocated stacks
    size_t rem = (size_ - guard) % (stacksize + guard);
    if(rem) {
      size_ += guard - rem;
    }
    // Now allocate our chunk.
    //std::cerr << "Allocating chunk of size " << size_ << "\n";
    int mmap_flags = MAP_SHARED;
#ifdef MAP_ANONYMOUS
    mmap_flags |= MAP_ANONYMOUS;
#else
    mmap_flags |= MAP_ANON;
#endif
    addr_ = (char*)mmap(0, size_, PROT_READ | PROT_WRITE | PROT_EXEC,
                        mmap_flags, -1, 0);
    if(addr_ == MAP_FAILED) {
      std::cerr << "Failed to mmap a region of size " << size_ << ": "
                << strerror(errno) << "\n";
      throw memoryerror("stackalloc::chunk:  Failed to mmap region.");
    }
    // and set protections on the pages between the stack chunks.
    const size_t stride = stacksize_ + guard;
    if(use_mprot_) {
      for(size_t offset = 0; offset < size_; offset += stride) {
        //std::cerr << "Setting mprotect at offset " << offset << "\n";
        mprotect(addr_+offset, guard, PROT_NONE);
      }
    }
  }

  //
  // Goodbye.
  //
  stackalloc::chunk::~chunk() {
    if(addr_) {
      munmap(addr_, size_);
    }
  }

} // end of namespace sstmac
