#
#   This file is part of SST/macroscale:
#                The macroscale architecture simulator from the SST suite.
#   Copyright (c) 2009 Sandia Corporation.
#   This software is distributed under the BSD License.
#   Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#   the U.S. Government retains certain rights in this software.
#   For more information, see the LICENSE file in the top
#   SST/macroscale directory.
#

import sys
import getopt
import sstmac
import math


# create networks, returns the sstmac::network too so we can print stuff out.

# mesh: n1xn2xn3 which n3 having PBC. routing currently doesn't use PBC.
def meshnet(latency, bandwidth, n1, n2 , n3):
    net = sstmac.composite.construct(
        sstmac.line.construct(n1),
        sstmac.composite.construct(
            sstmac.line.construct(n2),
            sstmac.ring.construct(n3)))
    return [net,
            sstmac.circuit.construct(sstmac.timestamp(latency), bandwidth, net)]

# gamma:
def gammanet(latency, bandwidth, degree, diameter):
    net = sstmac.gamma.construct(degree,diameter)
    return [net,
            sstmac.circuit.construct(sstmac.timestamp(latency), bandwidth, net)]

# fat: fat tree. links is 1/2 the links of the switch elements.
def fatnet(latency, bandwidth, levels, links):
    net = sstmac.fattree.construct(levels, links)
    return [net,
            sstmac.circuit.construct(sstmac.timestamp(latency), bandwidth, net)]

# torus: it says torus, but we don't use the PBC, so its really a mesh.
def torusnet(latency, bandwidth, n1, n2, n3):
    net = sstmac.torus.construct_3D(n1, n2, n3)
    return [net,
            sstmac.circuit.construct(sstmac.timestamp(latency), bandwidth, net)]

# full: full xbar switch.
def fullnet(latency, bandwidth):
    return [0,
            sstmac.fullnetwork.construct(sstmac.timestamp(latency), bandwidth)]

def pingpong(nodes, ppn, bandwidth, latency, verbose,networkname, randomize,iter,arrlen):
    """Run a simple mpi ping-pong"""

    if verbose:
        #sstmac.debug().to_stderr()
        print "nodes = ",nodes
        print "  ppn = ",ppn
        print "  lat = ",latency
        print "   bw = ",bandwidth
        print "  net = ",networkname
        print " rand = ",randomize
        print " iter = ",iter
        print "arrlen= ",arrlen


    mgr = sstmac.manager.construct()

    # for mesh like things, pick smallest cube that fits. TODO, trim
    # a dim to minimize size.
    nside = nodes ** (1.0/3.0)
    nside = int(nside+0.5)

    # log base fanout I think is the number of levels...
    nfanout = 12 # really 1/2 number of links of a xbar.
    nlevel = int( (math.log(nodes)/math.log(nfanout)) + 0.5)

    # todo add gamma.
    gamma_sizes = [ 2, (1,2),  3, (2,1),  4, (3,1),  5, (4,1),  6, (2,2),
                    6, (5,1),  7, (6,1),  8, (7,1),  9, (8,1), 12,(3,2),
                    20,(4,2),  24,(3,3),  30,(5,2),  42,(6,2),  56,(7,2),
                    60,(4,3),  72,(8,2), 120,(4,4), 120,(5,3), 210,(6,3),
                    336,(7,3),360,(5,4), 504,(8,3), 720,(5,5), 840,(6,4),
                    1680,(7,4), 2520,(6,5), 3024,(8,4), 5040,(6,6), 6720,(7,5),
                    15120,(8,5), 20160,(7,6), 40320,(7,7), 60480,(8,6),
                    181440,(8,7), 362880,(8,8)]

    # set nm_net and net and pretty string for network params.
    if networkname == "gamma":
        (gamma_degree, gamma_diameter) = (-1,-1)
        for i in range(len(gamma_sizes)/2):
            j = 2*i
            if verbose:
                print "%d gamma[%d] = " % (i, gamma_sizes[j]),
                print gamma_sizes[j+1]
            if gamma_sizes[j] >= nodes:
                (gamma_degree, gamma_diameter) = gamma_sizes[j+1]
                break
        if gamma_degree < 0:
            print "I need a better gamma table."
            return
        if verbose:
            print "I'm using %d %d" % (gamma_degree,gamma_diameter)
        params = "dgr=%d diam=%d" % (gamma_degree, gamma_diameter)
        [nm_net,net] = gammanet(latency, bandwidth, gamma_degree, gamma_diameter)
    elif networkname == "fat":
        [nm_net,net] = fatnet(latency, bandwidth, nlevel, nfanout)
        params = "l=%d k=%d" % (nlevel, nfanout)
    elif networkname == "torus":
        [nm_net,net] = torusnet(latency, bandwidth, nside, nside, nside)
        params = "%dx%dx%d" % (nside,nside,nside)
    elif networkname == "full":
        [nm_net,net] = fullnet(latency, bandwidth)
        params = ""
    elif networkname == "mesh":
        [nm_net,net]= meshnet(latency, bandwidth, nside, nside, nside)
        params = "%dx%dx%d" % (nside,nside,nside)
    else:
        print "Invalid network type:",network
        return
    # maxnode supported by network. maxnode 0 if full.
    if nm_net:
        if nm_net.n() != nm_net.nswitch(): 
            maxnode = nm_net.n() - nm_net.nswitch() # indirect network?
        else:
            maxnode = nm_net.n() # direct network?
    else:
        maxnode = 0

    starttime = 1.0 # start time for simulation

    mgr.set_interconnect(net)
    mgr.set_eventmanager(sstmac.eventmap.construct())
    mgr.add_nodes(sstmac.simplenode.construct(int(1e10)), nodes)
    job = sstmac.mpilaunch.construct(sstmac.mpipingpong.construct(iter,arrlen))
    job.set_scheduler(sstmac.scheduler.construct(32768, 10240000, False))

    if(randomize > 0):
        mgr.add_launcher(sstmac.timestamp(starttime), job, nodes, ppn,\
                         sstmac.randomindexing.construct(randomize))
    else:
        mgr.add_launcher(sstmac.timestamp(starttime), job, nodes, ppn,\
                         sstmac.blockindexing.construct())

    runtime = mgr.run().sec();
    if nm_net:
        params = "%d:%s" % (maxnode, params)
    else:
        if params:
            params = "inf:%s" % (params)

    if len(params) > 0: params = "%s," % (params);
            
    if randomize < 0:
        params = "%sblk" % (params)
    else:
        params = "%srnd" % (params)

    print '%4d %2d  %8d %2d  %6.5fs  %7.3fMB %7.3fus %-s(%-s)' % \
        (nodes,ppn,arrlen,iter,
         (runtime-starttime),
         (bandwidth/1000./1000.),(latency*1000.*1000.),networkname,params)

def main(args):
    """Parse arguments and run pingpong experiment."""
    (nodes, ppn, bandwidth, latency, verbose) \
        = (1000, 1, 147400000000, 0.0000013, False)
    (networkname, randomize) = ("full", -1)
    (iter, arrlen) = (10, 4)

    opts = getopt.gnu_getopt(args, "hvn:p:b:l:x:r:i:a:",
                             ['help','verbose','nodes=','ppn=','bandwidth=', 'latency=', 'network=','randomize=', 'iter=', 'arrlen='])
    if verbose:
        print opts
    for pair in opts[0]:
        if pair[0] == "-h" or pair[0] == "--help":
            print "sys.argv[0] [--help] [--nodes=<int>] [--ppn=<int>]",
            print "[--inputfile=<string>] ",
            print "[--bandwidth=<int>] [--latency=<float>]",
            print "[--network=<string>] [--randomize=<int>]",
            print "[--iter=<int>] [--arrlen=<int>]",
            return
        if pair[0] == "-v" or pair[0] == "--verbose":
            verbose = True
        if pair[0] == "-n" or pair[0] == "--nodes":
            nodes = int(pair[1])
        if pair[0] == "-p" or pair[0] == "--ppn":
            ppn = int(pair[1])
        if pair[0] == "-b" or pair[0] == "--bandwidth":
            bandwidth = float(pair[1])
        if pair[0] == "-l" or pair[0] == "--latency":
            latency = float(pair[1])
        if pair[0] == "-x" or pair[0] == "--network":
            networkname = pair[1]
        if pair[0] == "-r" or pair[0] == "--randomize":
            randomize = int(pair[1])
        if pair[0] == "-i" or pair[0] == "--iter":
            iter = int(pair[1])
        if pair[0] == "-a" or pair[0] == "--arrlen":
            arrlen = int(pair[1])

    pingpong(nodes, ppn, bandwidth, latency, verbose,networkname, randomize,iter,arrlen)

if __name__ == "__main__":
    main(sys.argv[1:])
